'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any, Union, List

import polars as pl

from pm4py.util import exec_utils, constants, xes_constants


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    SUBCASE_CONCAT_STR = "subcase_concat_str"


def apply(
    df: pl.LazyFrame,
    act1: Union[str, List[str]],
    act2: Union[str, List[str]],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Given a LazyFrame, filters all the subtraces going from an event with activity "act1" to an event with
    activity "act2"

    Parameters
    ----------------
    df
        LazyFrame
    act1
        First activity (or collection of activities)
    act2
        Second activity (or collection of activities)
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY => activity key
        - Parameters.CASE_ID_KEY => case id
        - Parameters.SUBCASE_CONCAT_STR => concatenator between the case id and the subtrace index in the filtered df

    Returns
    ----------------
    filtered_df
        LazyFrame with all the subtraces going from "act1" to "act2"
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    subcase_concat_str = exec_utils.get_param_value(
        Parameters.SUBCASE_CONCAT_STR, parameters, "##@@"
    )

    act1_list = [act1] if isinstance(act1, str) else act1
    act2_list = [act2] if isinstance(act2, str) else act2

    # Sort by case and timestamp
    df = df.sort([case_id_key, xes_constants.DEFAULT_TIMESTAMP_KEY])
    
    # Add row number within each case
    df = df.with_columns(
        pl.int_range(pl.len()).over(case_id_key).alias("_event_idx")
    )

    # Mark activities
    df = df.with_columns([
        pl.col(activity_key).is_in(act1_list).alias("_is_act1"),
        pl.col(activity_key).is_in(act2_list).alias("_is_act2")
    ])

    # For each act1, find the next act2 and create a segment identifier
    df = df.with_columns([
        # Mark positions where act1 occurs
        pl.when(pl.col("_is_act1"))
        .then(pl.col("_event_idx"))
        .otherwise(None)
        .alias("_act1_pos"),
        
        # Mark positions where act2 occurs  
        pl.when(pl.col("_is_act2"))
        .then(pl.col("_event_idx"))
        .otherwise(None)
        .alias("_act2_pos")
    ])

    # For each event, find the most recent act1 position before it
    df = df.with_columns(
        pl.col("_act1_pos")
        .fill_null(strategy="forward")
        .over(case_id_key)
        .alias("_current_act1")
    )

    # Mark events that are part of a valid segment (between act1 and act2 inclusive)
    df = df.with_columns(
        # An event is in a segment if:
        # 1. There's a current act1 before or at this position
        # 2. Either it's the act1 itself, or it's after act1 but before or at the next act2
        pl.when(
            pl.col("_current_act1").is_not_null() & 
            (pl.col("_event_idx") >= pl.col("_current_act1"))
        )
        .then(pl.col("_current_act1"))
        .otherwise(None)
        .alias("_segment_start")
    )

    # Find segment ends (act2 positions that have a preceding act1)
    df = df.with_columns(
        pl.when(
            pl.col("_is_act2") & 
            pl.col("_segment_start").is_not_null()
        )
        .then(pl.col("_event_idx"))
        .otherwise(None)
        .alias("_segment_end")
    )

    # For each segment end, we need to create a subcase
    # First, mark segment boundaries
    df = df.with_columns(
        pl.col("_segment_end")
        .fill_null(strategy="backward") 
        .over(case_id_key)
        .alias("_next_segment_end")
    )

    # Keep only events that are in complete segments
    df = df.filter(
        pl.col("_segment_start").is_not_null() &
        pl.col("_next_segment_end").is_not_null() &
        (pl.col("_event_idx") <= pl.col("_next_segment_end"))
    )

    # Create unique segment identifiers
    df = df.with_columns(
        # Create a unique identifier for each segment
        (pl.col("_segment_start").cast(pl.Utf8) + "_" + pl.col("_next_segment_end").cast(pl.Utf8))
        .alias("_segment_id")
    )
    
    # For each segment, check if it actually starts with act1 and ends with act2
    # Get the first and last activity of each segment
    segment_check = (
        df.group_by([case_id_key, "_segment_id"])
        .agg([
            pl.col(activity_key).first().alias("_first_act"),
            pl.col(activity_key).last().alias("_last_act"),
            pl.col("_segment_id").first().alias("_segment_id_check")
        ])
        .filter(
            pl.col("_first_act").is_in(act1_list) &
            pl.col("_last_act").is_in(act2_list)
        )
        .select([case_id_key, "_segment_id_check"])
        .rename({"_segment_id_check": "_segment_id"})
    )
    
    # Keep only segments that pass the check
    df = df.join(segment_check, on=[case_id_key, "_segment_id"], how="inner")

    # Assign subcase numbers
    df = df.with_columns(
        pl.col("_segment_id")
        .rank("dense")
        .over(case_id_key)
        .alias("_subcase_num")
    )

    # Update case ID with subcase information
    df = df.with_columns(
        pl.concat_str([
            pl.col(case_id_key).cast(pl.Utf8),
            pl.lit(subcase_concat_str),
            pl.col("_subcase_num").cast(pl.Utf8)
        ]).alias(case_id_key)
    )

    # Drop helper columns
    df = df.drop([
        "_event_idx", "_is_act1", "_is_act2", "_act1_pos", "_act2_pos",
        "_current_act1", "_segment_start", "_segment_end", "_next_segment_end",
        "_segment_id", "_subcase_num"
    ])

    return df


'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any, Tuple

import polars as pl

from pm4py.statistics.process_cube.polars.variants import classic
from pm4py.util import exec_utils


class Variants(Enum):
    CLASSIC = classic


def apply(
    feature_table: pl.LazyFrame | pl.DataFrame,
    x_col: str,
    y_col: str,
    agg_col: str,
    variant=Variants.CLASSIC,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Tuple[pl.DataFrame, Dict[Any, Any]]:
    """Applies the selected process cube variant using Polars data structures."""

    return exec_utils.get_variant(variant).apply(
        feature_table, x_col, y_col, agg_col, parameters=parameters
    )



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.algo.concept_drift.variants import bose
from pm4py.util import exec_utils
import pandas as pd
from pm4py.objects.log.obj import EventLog
from typing import Union, Dict, Any, Optional, Tuple, List


class Variants(Enum):
    BOSE = bose


def apply(log: Union[EventLog, pd.DataFrame], variant=Variants.BOSE, parameters: Optional[Dict[Any, Any]] = None) -> \
Tuple[List[pd.DataFrame], List[int], List[float]]:
    """
    Parameters
    --------------
    log
        Event log or Pandas dataframe
    variant
        Variant of the algorithm (available: Variants.BOSE)
    parameters
        Variant-specific parameters

    Returns
    ---------------
    returned_sublogs : List[EventLog]
        A list of sub-logs, where each sub-log is an EventLog object representing the cumulative segment of the original event log from the start up to each detected change point (and the final sub-log up to the end). Note: Due to a potential implementation issue, these sub-logs are not segments between change points but rather cumulative logs up to each change point.
    change_timestamps : List[float]
        A list of timestamps where concept drifts are detected. Each timestamp corresponds to the start time of the first trace in the sub-log where a change point occurs, based on case start timestamps.
    p_values : List[float]
        A list of p-values associated with each detected change point, indicating the statistical significance of the drift (lower values suggest stronger evidence of a change).
    """
    return exec_utils.get_variant(variant).apply(log, parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.streaming.algo.discovery.dfg.variants import frequency
from enum import Enum
from pm4py.util import exec_utils


class Variants(Enum):
    FREQUENCY = frequency


DEFAULT_VARIANT = Variants.FREQUENCY


def apply(variant=DEFAULT_VARIANT, parameters=None):
    """
    Discovers a DFG from an event stream

    Parameters
    --------------
    variant
        Variant of the algorithm (default: Variants.FREQUENCY)

    Returns
    --------------
    stream_dfg_obj
        Streaming DFG discovery object
    """
    if parameters is None:
        parameters = {}

    return exec_utils.get_variant(variant).apply(parameters=parameters)



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.visualization.graphs.variants import (
    cases,
    attributes,
    dates,
    barplot,
)
from pm4py.visualization.graphs.util.common import (
    save,
    view,
    matplotlib_view,
    serialize,
)
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, List


class Variants(Enum):
    CASES = cases
    ATTRIBUTES = attributes
    DATES = dates
    BARPLOT = barplot


DEFAULT_VARIANT = Variants.CASES


def apply(
    x: List[float],
    y: List[float],
    parameters: Optional[Dict[Any, Any]] = None,
    variant=DEFAULT_VARIANT,
) -> str:
    """
    Method to plot (non-logarithmic way) the graph with axis values contained in x and y

    Parameters
    ------------
    x
        Values for x-axis
    y
        Values for y-axis
    parameters
        Parameters of the algorithm, including:
            Parameters.FORMAT -> Format of the target image
            Parameters.TITLE -> Title of the image
    variant
        Variant of the algorithm to apply, including:
            - Variants.CASES
            - Variants.ATTRIBUTES
            - Variants.DATES
            - Variants.BARPLOT

    Returns
    ------------
    temp_file_name
        Representation temporary file name
    """
    return exec_utils.get_variant(variant).apply_plot(
        x, y, parameters=parameters
    )


def apply_plot(
    x: List[float],
    y: List[float],
    parameters: Optional[Dict[Any, Any]] = None,
    variant=DEFAULT_VARIANT,
) -> str:
    """
    Method to plot (non-logarithmic way) the graph with axis values contained in x and y

    Parameters
    ------------
    x
        Values for x-axis
    y
        Values for y-axis
    parameters
        Parameters of the algorithm, including:
            Parameters.FORMAT -> Format of the target image
            Parameters.TITLE -> Title of the image
    variant
        Variant of the algorithm to apply, including:
            - Variants.CASES
            - Variants.ATTRIBUTES
            - Variants.DATES
            - Variants.BARPLOT

    Returns
    ------------
    temp_file_name
        Representation temporary file name
    """
    return exec_utils.get_variant(variant).apply_plot(
        x, y, parameters=parameters
    )


def apply_semilogx(
    x: List[float],
    y: List[float],
    parameters: Optional[Dict[Any, Any]] = None,
    variant=DEFAULT_VARIANT,
) -> str:
    """
    Method to plot (semi-logarithmic way) the graph with axis values contained in x and y

    Parameters
    ------------
    x
        Values for x-axis
    y
        Values for y-axis
    parameters
        Parameters of the algorithm, including:
            Parameters.FORMAT -> Format of the target image
            Parameters.TITLE -> Title of the image
    variant
        Variant of the algorithm to apply, including:
            - Variants.CASES
            - Variants.ATTRIBUTES
            - Variants.DATES
            - Variants.BARPLOT

    Returns
    ------------
    temp_file_name
        Representation temporary file name
    """
    return exec_utils.get_variant(variant).apply_semilogx(
        x, y, parameters=parameters
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any, Tuple, Collection, Set, List
from enum import Enum
from pm4py.util import exec_utils, constants
from pm4py.objects.ocel import constants as ocel_constants
from pm4py.util.business_hours import BusinessHours
import numpy as np
import datetime


class Parameters(Enum):
    EVENT_ID = ocel_constants.PARAM_EVENT_ID
    OBJECT_ID = ocel_constants.PARAM_OBJECT_ID
    OBJECT_TYPE = ocel_constants.PARAM_OBJECT_TYPE
    EVENT_ACTIVITY = ocel_constants.PARAM_EVENT_ACTIVITY
    EVENT_TIMESTAMP = ocel_constants.PARAM_EVENT_TIMESTAMP
    BUSINESS_HOURS = "business_hours"
    BUSINESS_HOUR_SLOTS = "business_hour_slots"
    WORKCALENDAR = "workcalendar"


def performance_calculation_ocel_aggregation(
    ocel: OCEL,
    aggregation: Dict[str, Dict[Tuple[str, str], Set[Any]]],
    parameters: Optional[Dict[Any, Any]] = None,
) -> Dict[str, Dict[Tuple[str, str], List[float]]]:
    """
    Calculates the performance based on one of the following aggregations:
    - aggregate_ev_couples
    - aggregate_total_objects

    Parameters
    ----------------
    ocel
        Object-centric event log
    aggregation
        Aggregation calculated using one of the aforementioned methods
    parameters
        Parameters of the algorithm, including:
        - Parameters.EVENT_ID => the event identifier
        - Parameters.EVENT_TIMESTAMP => the timestamp
        - Parameters.BUSINESS_HOURS => enables/disables the business hours
        - Parameters.BUSINESS_HOURS_SLOTS =>
        work schedule of the company, provided as a list of tuples where each tuple represents one time slot of business
        hours. One slot i.e. one tuple consists of one start and one end time given in seconds since week start, e.g.
        [
            (7 * 60 * 60, 17 * 60 * 60),
            ((24 + 7) * 60 * 60, (24 + 12) * 60 * 60),
            ((24 + 13) * 60 * 60, (24 + 17) * 60 * 60),
        ]
        meaning that business hours are Mondays 07:00 - 17:00 and Tuesdays 07:00 - 12:00 and 13:00 - 17:00

    Returns
    ----------------
    edges_performance
        For each object type, associate a dictionary where to each activity couple
        all the times between the activities are recorded.
    """
    if parameters is None:
        parameters = {}

    event_id = exec_utils.get_param_value(
        Parameters.EVENT_ID, parameters, ocel.event_id_column
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.EVENT_TIMESTAMP, parameters, ocel.event_timestamp
    )
    timestamps = (
        ocel.events.groupby(event_id)[timestamp_key].agg(list).to_dict()
    )
    timestamps = {x: y[0] for x, y in timestamps.items()}

    business_hours = exec_utils.get_param_value(
        Parameters.BUSINESS_HOURS, parameters, False
    )
    business_hours_slots = exec_utils.get_param_value(
        Parameters.BUSINESS_HOUR_SLOTS,
        parameters,
        constants.DEFAULT_BUSINESS_HOUR_SLOTS,
    )

    workcalendar = exec_utils.get_param_value(
        Parameters.WORKCALENDAR,
        parameters,
        constants.DEFAULT_BUSINESS_HOURS_WORKCALENDAR,
    )

    ret = {}

    for ot in aggregation:
        ret[ot] = {}
        for act in aggregation[ot]:
            ret[ot][act] = []
            for el in aggregation[ot][act]:
                if business_hours:
                    bh = BusinessHours(
                        timestamps[el[0]],
                        timestamps[el[1]],
                        business_hour_slots=business_hours_slots,
                        workcalendar=workcalendar,
                    )
                    diff = bh.get_seconds()
                else:
                    timedelta = timestamps[el[1]] - timestamps[el[0]]
                    diff = 0
                    if isinstance(timedelta, np.timedelta64):
                        diff = timedelta / np.timedelta64(1, "s")
                    elif isinstance(timedelta, datetime.timedelta):
                        diff = timedelta.total_seconds()
                ret[ot][act].append(diff)
            ret[ot][act] = sorted(ret[ot][act])

    return ret


def aggregate_ev_couples(
    edges: Dict[str, Dict[Tuple[str, str], Collection[Any]]]
) -> Dict[str, Dict[Tuple[str, str], Set[Any]]]:
    """
    Performs an aggregation of the occurrences of a given edge on the couple of events (source event, target event).

    Parameters
    -------------------
    edges
        Edges calculated using the find_associations_per_edge function

    Returns
    -------------------
    aggregation
        A dictionary associating to each object type another dictionary where to each edge (activity couple) all the
        couples of related events are associated.
    """
    ret = {}
    for ot in edges:
        ret[ot] = {}
        for act in edges[ot]:
            ret[ot][act] = set((x[0], x[1]) for x in edges[ot][act])
    return ret


def aggregate_unique_objects(
    edges: Dict[str, Dict[Tuple[str, str], Collection[Any]]]
) -> Dict[str, Dict[Tuple[str, str], Set[Any]]]:
    """
    Performs an aggregation of the occurrences of a given edge in the involved object.

    Parameters
    -------------------
    edges
        Edges calculated using the find_associations_per_edge function

    Returns
    -------------------
    aggregation
        A dictionary associating to each object type another dictionary where to each edge (activity couple) all the
        involved objects are associated.
    """
    ret = {}
    for ot in edges:
        ret[ot] = {}
        for act in edges[ot]:
            ret[ot][act] = set(x[2] for x in edges[ot][act])
    return ret


def aggregate_total_objects(
    edges: Dict[str, Dict[Tuple[str, str], Collection[Any]]]
) -> Dict[str, Dict[Tuple[str, str], Set[Any]]]:
    """
    Performs an aggregation of the occurrences of a given edge on the triple (source event, target event, object).

    Parameters
    -------------------
    edges
        Edges calculated using the find_associations_per_edge function

    Returns
    -------------------
    aggregation
        A dictionary associating to each object type another dictionary where to each edge (activity couple) all the
        triples (source event, target event, object) are associated.
    """
    ret = {}
    for ot in edges:
        ret[ot] = {}
        for act in edges[ot]:
            ret[ot][act] = set(edges[ot][act])
    return ret


def find_associations_per_edge(
    ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None
) -> Dict[str, Dict[Tuple[str, str], Collection[Any]]]:
    """
    Finds all the occurrences of a given edge (activity couple), expressed as triples (source event, target event, object ID).

    Parameters
    -------------------
    ocel
        Object-centric event log
    parameters
        Parameters of the algorithm, including:
        - Parameters.EVENT_ACTIVITY => the activity
        - Parameters.EVENT_ID => the event identifier
        - Parameters.OBJECT_ID => the object identifier
        - Parameters.OBJECT_TYPE => the object type

    Returns
    ------------------
    edges
        A dictionary associating to each object type a dictionary where to each edge (activity couple) the list of triples (source event, target event, object ID)
        is associated.
    """
    if parameters is None:
        parameters = {}

    event_activity = exec_utils.get_param_value(
        Parameters.EVENT_ACTIVITY, parameters, ocel.event_activity
    )
    event_id = exec_utils.get_param_value(
        Parameters.EVENT_ID, parameters, ocel.event_id_column
    )
    object_id = exec_utils.get_param_value(
        Parameters.OBJECT_ID, parameters, ocel.object_id_column
    )
    object_type = exec_utils.get_param_value(
        Parameters.OBJECT_TYPE, parameters, ocel.object_type_column
    )

    identifiers = ocel.events[event_id].to_numpy().tolist()
    activities = (
        ocel.events.groupby(event_id)[event_activity].agg(list).to_dict()
    )
    activities = {x: y[0] for x, y in activities.items()}

    omap = ocel.relations.groupby(event_id)[object_id].agg(list).to_dict()
    objtypes = ocel.objects.groupby(object_id)[object_type].agg(list).to_dict()
    objtypes = {x: y[0] for x, y in objtypes.items()}

    history = {}
    edges = {}

    for evid in identifiers:
        if evid in omap:
            for obj in omap[evid]:
                if obj in history:
                    objtype = objtypes[obj]
                    if objtype not in edges:
                        edges[objtype] = {}
                    previd = history[obj]
                    acttup = (activities[previd], activities[evid])
                    if acttup not in edges[objtype]:
                        edges[objtype][acttup] = list()
                    edges[objtype][acttup].append((previd, evid, obj))
                history[obj] = evid

    return edges



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from typing import List, TypeVar, Tuple, Optional, Dict, Any

from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructure,
    IMDataStructureUVCL,
)
from pm4py.algo.discovery.inductive.fall_through.abc import FallThrough
from pm4py.algo.discovery.inductive.fall_through.activity_concurrent import (
    ActivityConcurrentUVCL,
)
from pm4py.algo.discovery.inductive.fall_through.activity_once_per_trace import (
    ActivityOncePerTraceUVCL, )
from pm4py.algo.discovery.inductive.fall_through.empty_traces import (
    EmptyTracesUVCL,
    EmptyTracesDFG,
)
from pm4py.algo.discovery.inductive.fall_through.flower import (
    FlowerModelUVCL,
    FlowerModelDFG,
)
from pm4py.algo.discovery.inductive.fall_through.strict_tau_loop import (
    StrictTauLoopUVCL,
)
from pm4py.algo.discovery.inductive.fall_through.tau_loop import TauLoopUVCL
from pm4py.algo.discovery.inductive.variants.instances import IMInstance
from pm4py.objects.process_tree.obj import ProcessTree
from pm4py.util import exec_utils
from enum import Enum


T = TypeVar("T", bound=IMDataStructure)
S = TypeVar("S", bound=FallThrough)


class Parameters(Enum):
    DISABLE_FALLTHROUGHS = "disable_fallthroughs"


class FallThroughFactory:

    @classmethod
    def get_fall_throughs(
        cls,
        obj: T,
        inst: IMInstance,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[S]:
        if parameters is None:
            parameters = {}

        disable_fallthroughs = exec_utils.get_param_value(
            Parameters.DISABLE_FALLTHROUGHS, parameters, False
        )

        if inst is IMInstance.IM or inst is IMInstance.IMf:
            if type(obj) is IMDataStructureUVCL:
                if disable_fallthroughs:
                    return [EmptyTracesUVCL, FlowerModelUVCL]
                else:
                    return [
                        EmptyTracesUVCL,
                        ActivityOncePerTraceUVCL,
                        ActivityConcurrentUVCL,
                        StrictTauLoopUVCL,
                        TauLoopUVCL,
                        FlowerModelUVCL,
                    ]
        if inst is IMInstance.IMd:
            if disable_fallthroughs:
                return [EmptyTracesDFG, FlowerModelDFG]
            else:
                return [EmptyTracesDFG, FlowerModelDFG]
        return list()

    @classmethod
    def fall_through(
        cls,
        obj: T,
        inst: IMInstance,
        pool,
        manager,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Tuple[ProcessTree, List[T]]:
        for f in FallThroughFactory.get_fall_throughs(obj, inst, parameters):
            r = f.apply(obj, pool, manager, parameters)
            if r is not None:
                return r
        return None



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from re import escape


def translate_infix_to_regex(infix):
    regex = "^"
    for i, act in enumerate(infix):
        is_last_activity = i == (len(infix) - 1)
        if act == "...":
            if is_last_activity:
                regex = f"{regex[:-1]}(,[^,]*)*"
            else:
                regex = f"{regex}([^,]*,)*"
        else:
            if act:
                act = escape(act)

            if is_last_activity:
                regex = f"{regex}{act}"
            else:
                regex = f"{regex}{act},"

    regex = f"{regex}$"
    return regex



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from typing import Dict, List, Tuple
from pm4py.objects.petri_net.stochastic.obj import StochasticPetriNet


class StochasticArcWeightNet(StochasticPetriNet):
    """
    Petri nets with stochastic arc weights. Arcs are assumed to desribe distributions of token consumption/production.
    We utilize the weight attribute that is defined in the Petri net based class to store the distribution.
    """

    class Arc(StochasticPetriNet.Arc):

        def __init__(self, source, target, weight={1: 1.0}, properties=None):
            super().__init__(source, target, weight, properties)

        def __get_weight_distribution(self) -> Dict[int, float]:
            return self.weight

        weight_distribution = property(__get_weight_distribution)

    Binding = List[Tuple[Arc, int]]



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import importlib.util


def apply(input_path, validation_path, parameters=None):
    if not importlib.util.find_spec("lxml"):
        raise Exception(
            "please install lxml in order to validate an XMLOCEL file."
        )

    import lxml.etree

    if parameters is None:
        parameters = {}

    xml_file = lxml.etree.parse(input_path)
    xml_validator = lxml.etree.XMLSchema(file=validation_path)
    is_valid = xml_validator.validate(xml_file)
    return is_valid



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.bpmn.obj import BPMN
from pm4py.util import constants, exec_utils
from enum import Enum


class Parameters(Enum):
    ENCODING = "encoding"


class Counts:
    def __init__(self):
        self.number_processes = 0


def parse_element(
    bpmn_graph,
    counts,
    curr_el,
    parents,
    incoming_dict,
    outgoing_dict,
    nodes_dict,
    nodes_bounds,
    flow_info,
    process=None,
    node=None,
    bpmn_element=None,
    flow=None,
    rec_depth=0,
):
    """
    Parses a BPMN element from the XML file
    """
    layout = bpmn_graph.get_layout()
    tag = curr_el.tag.lower()

    if curr_el.get("id") is not None:
        if tag.endswith("collaboration"):
            collaboration_id = curr_el.get("id")
            collaboration = BPMN.Collaboration(
                id=collaboration_id, process=collaboration_id
            )
            bpmn_graph.add_node(collaboration)
            nodes_dict[collaboration_id] = collaboration
        elif tag.endswith("participant"):
            participant_id = curr_el.get("id")
            name = (
                curr_el.get("name").replace("\r", "").replace("\n", "")
                if "name" in curr_el.attrib
                else ""
            )
            process_ref = curr_el.get("processRef")
            participant = BPMN.Participant(
                id=participant_id,
                name=name,
                process=None,
                process_ref=process_ref,
            )
            bpmn_graph.add_node(participant)
            nodes_dict[participant_id] = participant
        elif tag.endswith("textannotation"):
            annotation_id = curr_el.get("id")
            text = ""
            for child in curr_el:
                text = child.text
            annotation = BPMN.TextAnnotation(
                id=annotation_id, text=text, process=process
            )
            bpmn_graph.add_node(annotation)
            nodes_dict[annotation_id] = annotation
        elif tag.endswith("subprocess"):  # subprocess invocation
            name = (
                curr_el.get("name").replace("\r", "").replace("\n", "")
                if "name" in curr_el.attrib
                else ""
            )
            subprocess = BPMN.SubProcess(
                id=curr_el.get("id"),
                name=name,
                process=process,
                depth=rec_depth,
            )
            bpmn_graph.add_node(subprocess)
            node = subprocess
            process = curr_el.get("id")
            nodes_dict[process] = node
        elif tag.endswith("process"):  # process of the current subtree
            process = curr_el.get("id")
            name = (
                curr_el.get("name").replace("\r", "").replace("\n", "")
                if "name" in curr_el.attrib
                else ""
            )
            bpmn_graph.set_process_id(process)
            bpmn_graph.set_name(name)
        elif tag.endswith(
            "shape"
        ):  # shape of a node, contains x,y,width,height information
            bpmn_element = curr_el.get("bpmnElement")
        elif tag.endswith("task"):  # simple task object
            id = curr_el.get("id")
            name = (
                curr_el.get("name").replace("\r", "").replace("\n", "")
                if "name" in curr_el.attrib
                else ""
            )
            # this_type = str(curr_el.tag)
            # this_type = this_type[this_type.index("}") + 1:]
            if tag.endswith("usertask"):
                task = BPMN.UserTask(id=id, name=name, process=process)
            elif tag.endswith("sendtask"):
                task = BPMN.SendTask(id=id, name=name, process=process)
            else:
                task = BPMN.Task(id=id, name=name, process=process)
            bpmn_graph.add_node(task)
            node = task
            nodes_dict[id] = node
        elif tag.endswith(
            "startevent"
        ):  # start node starting the (sub)process
            id = curr_el.get("id")
            name = (
                curr_el.get("name").replace("\r", " ").replace("\n", " ")
                if "name" in curr_el.attrib
                else ""
            )
            event_definitions = [
                child.tag.lower().replace("eventdefinition", "")
                for child in curr_el
                if child.tag.lower().endswith("eventdefinition")
            ]
            if len(event_definitions) > 0:
                event_type = event_definitions[0]
                if event_type.endswith("message"):
                    start_event = BPMN.MessageStartEvent(
                        id=curr_el.get("id"), name=name, process=process
                    )
                else:  # TODO: expand functionality, support more start event types
                    start_event = BPMN.NormalStartEvent(
                        id=curr_el.get("id"), name=name, process=process
                    )
            else:
                start_event = BPMN.NormalStartEvent(
                    id=curr_el.get("id"), name=name, process=process
                )
            bpmn_graph.add_node(start_event)
            node = start_event
            nodes_dict[id] = node
        elif tag.endswith("endevent"):  # end node ending the (sub)process
            id = curr_el.get("id")
            name = (
                curr_el.get("name").replace("\r", " ").replace("\n", " ")
                if "name" in curr_el.attrib
                else ""
            )
            event_definitions = [
                child.tag.lower().replace("eventdefinition", "")
                for child in curr_el
                if child.tag.lower().endswith("eventdefinition")
            ]
            if len(event_definitions) > 0:
                event_type = event_definitions[0]
                if event_type.endswith("message"):
                    end_event = BPMN.MessageEndEvent(
                        id=curr_el.get("id"), name=name, process=process
                    )
                elif event_type.endswith("terminate"):
                    end_event = BPMN.TerminateEndEvent(
                        id=curr_el.get("id"), name=name, process=process
                    )
                elif event_type.endswith("error"):
                    end_event = BPMN.ErrorEndEvent(
                        id=curr_el.get("id"), name=name, process=process
                    )
                elif event_type.endswith("cancel"):
                    end_event = BPMN.CancelEndEvent(
                        id=curr_el.get("id"), name=name, process=process
                    )
                else:  # TODO: expand functionality, support more start event types
                    end_event = BPMN.NormalEndEvent(
                        id=curr_el.get("id"), name=name, process=process
                    )
            else:
                end_event = BPMN.NormalEndEvent(
                    id=curr_el.get("id"), name=name, process=process
                )
            bpmn_graph.add_node(end_event)
            node = end_event
            nodes_dict[id] = node
        elif tag.endswith(
            "intermediatecatchevent"
        ):  # intermediate event that happens (externally) and can be catched
            id = curr_el.get("id")
            name = (
                curr_el.get("name").replace("\r", " ").replace("\n", " ")
                if "name" in curr_el.attrib
                else ""
            )
            event_definitions = [
                child.tag.lower().replace("eventdefinition", "")
                for child in curr_el
                if child.tag.lower().endswith("eventdefinition")
            ]
            if len(event_definitions) > 0:
                event_type = event_definitions[0]
                if event_type.endswith("message"):
                    intermediate_catch_event = (
                        BPMN.MessageIntermediateCatchEvent(
                            id=curr_el.get("id"), name=name, process=process
                        )
                    )
                elif event_type.endswith("error"):
                    intermediate_catch_event = (
                        BPMN.ErrorIntermediateCatchEvent(
                            id=curr_el.get("id"), name=name, process=process
                        )
                    )
                elif event_type.endswith("cancel"):
                    intermediate_catch_event = (
                        BPMN.CancelIntermediateCatchEvent(
                            id=curr_el.get("id"), name=name, process=process
                        )
                    )
                else:
                    intermediate_catch_event = BPMN.IntermediateCatchEvent(
                        id=curr_el.get("id"), name=name, process=process
                    )
            else:
                intermediate_catch_event = BPMN.IntermediateCatchEvent(
                    id=curr_el.get("id"), name=name, process=process
                )
            bpmn_graph.add_node(intermediate_catch_event)
            node = intermediate_catch_event
            nodes_dict[id] = node
        elif tag.endswith(
            "intermediatethrowevent"
        ):  # intermediate event that is activated through the (sub)process
            id = curr_el.get("id")
            name = (
                curr_el.get("name").replace("\r", " ").replace("\n", " ")
                if "name" in curr_el.attrib
                else ""
            )
            event_definitions = [
                child.tag.lower().replace("eventdefinition", "")
                for child in curr_el
                if child.tag.lower().endswith("eventdefinition")
            ]
            if len(event_definitions) > 0:
                event_type = event_definitions[0]
                if event_type.endswith("message"):
                    intermediate_throw_event = (
                        BPMN.MessageIntermediateThrowEvent(
                            id=curr_el.get("id"), name=name, process=process
                        )
                    )
                else:
                    intermediate_throw_event = (
                        BPMN.NormalIntermediateThrowEvent(
                            id=curr_el.get("id"), name=name, process=process
                        )
                    )
            else:
                intermediate_throw_event = BPMN.NormalIntermediateThrowEvent(
                    id=curr_el.get("id"), name=name, process=process
                )
            bpmn_graph.add_node(intermediate_throw_event)
            node = intermediate_throw_event
            nodes_dict[id] = node
        elif tag.endswith("boundaryevent"):
            id = curr_el.get("id")
            ref_activity = curr_el.get("attachedToRef")
            name = (
                curr_el.get("name").replace("\r", " ").replace("\n", " ")
                if "name" in curr_el.attrib
                else ""
            )
            event_definitions = [
                child.tag.lower().replace("eventdefinition", "")
                for child in curr_el
                if child.tag.lower().endswith("eventdefinition")
            ]
            if len(event_definitions) > 0:
                event_type = event_definitions[0]
                if event_type.endswith("message"):
                    boundary_event = BPMN.MessageBoundaryEvent(
                        id=curr_el.get("id"),
                        name=name,
                        process=process,
                        activity=ref_activity,
                    )
                elif event_type.endswith("error"):
                    boundary_event = BPMN.ErrorBoundaryEvent(
                        id=curr_el.get("id"),
                        name=name,
                        process=process,
                        activity=ref_activity,
                    )
                elif event_type.endswith("cancel"):
                    boundary_event = BPMN.CancelBoundaryEvent(
                        id=curr_el.get("id"),
                        name=name,
                        process=process,
                        activity=ref_activity,
                    )
                else:
                    boundary_event = BPMN.BoundaryEvent(
                        id=curr_el.get("id"),
                        name=name,
                        process=process,
                        activity=ref_activity,
                    )
            else:
                boundary_event = BPMN.BoundaryEvent(
                    id=curr_el.get("id"),
                    name=name,
                    process=process,
                    activity=ref_activity,
                )
            bpmn_graph.add_node(boundary_event)
            node = boundary_event
            nodes_dict[id] = node
        elif tag.endswith("edge"):  # related to the x, y information of an arc
            bpmnElement = curr_el.get("bpmnElement")
            flow = bpmnElement
        elif tag.endswith("exclusivegateway"):
            id = curr_el.get("id")
            name = (
                curr_el.get("name").replace("\r", "").replace("\n", "")
                if "name" in curr_el.attrib
                else ""
            )
            try:
                direction = BPMN.Gateway.Direction[
                    curr_el.get("gatewayDirection").upper()
                ]
                exclusive_gateway = BPMN.ExclusiveGateway(
                    id=curr_el.get("id"),
                    name=name,
                    gateway_direction=direction,
                    process=process,
                )
            except BaseException:
                exclusive_gateway = BPMN.ExclusiveGateway(
                    id=curr_el.get("id"),
                    name=name,
                    gateway_direction=BPMN.Gateway.Direction.UNSPECIFIED,
                    process=process,
                )
            bpmn_graph.add_node(exclusive_gateway)
            node = exclusive_gateway
            nodes_dict[id] = node
        elif tag.endswith("parallelgateway"):
            id = curr_el.get("id")
            name = (
                curr_el.get("name").replace("\r", "").replace("\n", "")
                if "name" in curr_el.attrib
                else ""
            )
            try:
                direction = BPMN.Gateway.Direction[
                    curr_el.get("gatewayDirection").upper()
                ]
                parallel_gateway = BPMN.ParallelGateway(
                    id=curr_el.get("id"),
                    name=name,
                    gateway_direction=direction,
                    process=process,
                )
            except BaseException:
                parallel_gateway = BPMN.ParallelGateway(
                    id=curr_el.get("id"),
                    name=name,
                    gateway_direction=BPMN.Gateway.Direction.UNSPECIFIED,
                    process=process,
                )
            bpmn_graph.add_node(parallel_gateway)
            node = parallel_gateway
            nodes_dict[id] = node
        elif tag.endswith("inclusivegateway"):
            id = curr_el.get("id")
            name = (
                curr_el.get("name").replace("\r", "").replace("\n", "")
                if "name" in curr_el.attrib
                else ""
            )
            try:
                direction = BPMN.Gateway.Direction[
                    curr_el.get("gatewayDirection").upper()
                ]
                inclusive_gateway = BPMN.InclusiveGateway(
                    id=curr_el.get("id"),
                    name=name,
                    gateway_direction=direction,
                    process=process,
                )
            except BaseException:
                inclusive_gateway = BPMN.InclusiveGateway(
                    id=curr_el.get("id"),
                    name=name,
                    gateway_direction=BPMN.Gateway.Direction.UNSPECIFIED,
                    process=process,
                )
            bpmn_graph.add_node(inclusive_gateway)
            node = inclusive_gateway
            nodes_dict[id] = node
        elif tag.endswith("eventbasedgateway"):
            id = curr_el.get("id")
            name = (
                curr_el.get("name").replace("\r", "").replace("\n", "")
                if "name" in curr_el.attrib
                else ""
            )
            try:
                direction = BPMN.Gateway.Direction[
                    curr_el.get("gatewayDirection").upper()
                ]
                event_based_gateway = BPMN.EventBasedGateway(
                    id=curr_el.get("id"),
                    name=name,
                    gateway_direction=direction,
                    process=process,
                )
            except BaseException:
                event_based_gateway = BPMN.EventBasedGateway(
                    id=curr_el.get("id"),
                    name=name,
                    gateway_direction=BPMN.Gateway.Direction.UNSPECIFIED,
                    process=process,
                )
            bpmn_graph.add_node(event_based_gateway)
            node = event_based_gateway
            nodes_dict[id] = node
        # FIXED: Moved the flow handling inside the id block
        elif (
            tag.endswith("sequenceflow")
            or tag.endswith("messageflow")
            or tag.endswith("association")
        ):
            seq_flow_id = curr_el.get("id")
            source_ref = curr_el.get("sourceRef")
            target_ref = curr_el.get("targetRef")
            name = (
                curr_el.get("name").replace("\r", "").replace("\n", "")
                if "name" in curr_el.attrib
                else ""
            )
            if source_ref is not None and target_ref is not None:
                incoming_dict[seq_flow_id] = (
                    target_ref,
                    process,
                    tag,
                    name,
                    parents,
                )
                outgoing_dict[seq_flow_id] = (
                    source_ref,
                    process,
                    tag,
                    name,
                    parents,
                )
    elif tag.endswith("incoming"):  # incoming flow of a node
        name = (
            curr_el.get("name").replace("\r", "").replace("\n", "")
            if "name" in curr_el.attrib
            else ""
        )
        if node is not None:
            if curr_el.text.strip() not in incoming_dict:
                incoming_dict[curr_el.text.strip()] = (
                    node,
                    process,
                    tag,
                    name,
                    parents,
                )
    elif tag.endswith("outgoing"):  # outgoing flow of a node
        name = (
            curr_el.get("name").replace("\r", "").replace("\n", "")
            if "name" in curr_el.attrib
            else ""
        )
        if node is not None:
            if curr_el.text.strip() not in outgoing_dict:
                outgoing_dict[curr_el.text.strip()] = (
                    node,
                    process,
                    tag,
                    name,
                    parents,
                )
    elif tag.endswith(
        "waypoint"
    ):  # contains information of x, y values of an edge
        if flow is not None:
            x = float(curr_el.get("x"))
            y = float(curr_el.get("y"))
            if flow not in flow_info:
                flow_info[flow] = []
            flow_info[flow].append((x, y))
    elif tag.endswith(
        "label"
    ):  # label of a node, mostly at the end of a shape object
        bpmn_element = None
    elif tag.endswith(
        "bounds"
    ):  # contains information of width, height, x, y of a node
        if bpmn_element is not None:
            x = float(curr_el.get("x"))
            y = float(curr_el.get("y"))
            width = float(curr_el.get("width"))
            height = float(curr_el.get("height"))
            nodes_bounds[bpmn_element] = {
                "x": x,
                "y": y,
                "width": width,
                "height": height,
            }

    for child in curr_el:
        bpmn_graph = parse_element(
            bpmn_graph,
            counts,
            child,
            list(parents) + [child],
            incoming_dict,
            outgoing_dict,
            nodes_dict,
            nodes_bounds,
            flow_info,
            process=process,
            node=node,
            bpmn_element=bpmn_element,
            flow=flow,
            rec_depth=rec_depth + 1,
        )
    # afterprocessing when the xml tree has been recursively parsed already
    if rec_depth == 0:
        # bpmn_graph.set_process_id(process)
        for seq_flow_id in incoming_dict:
            if incoming_dict[seq_flow_id][0] in nodes_dict:
                incoming_dict[seq_flow_id] = (
                    nodes_dict[incoming_dict[seq_flow_id][0]],
                    incoming_dict[seq_flow_id][1],
                    incoming_dict[seq_flow_id][2],
                    incoming_dict[seq_flow_id][3],
                    incoming_dict[seq_flow_id][4],
                )
        for seq_flow_id in outgoing_dict:
            if outgoing_dict[seq_flow_id][0] in nodes_dict:
                outgoing_dict[seq_flow_id] = (
                nodes_dict[outgoing_dict[seq_flow_id][0]], outgoing_dict[seq_flow_id][1], outgoing_dict[seq_flow_id][2],
                outgoing_dict[seq_flow_id][3], outgoing_dict[seq_flow_id][4])

        # also supports flows without waypoints
        flows_without_waypoints = set(flow_info).union(
            set(outgoing_dict).intersection(set(incoming_dict))
        )
        for flow_id in flows_without_waypoints:
            flow_info[flow_id] = []

        for flow_id in flow_info:
            if flow_id in outgoing_dict and flow_id in incoming_dict:
                flow = None
                flow_type = outgoing_dict[flow_id][2]

                if flow_type.endswith("messageflow"):
                    collaboration_id = None
                    for par in outgoing_dict[flow_id][4]:
                        par_tag = str(par.tag).lower()
                        if par_tag.endswith("collaboration"):
                            collaboration_id = par.get("id")
                    flow = BPMN.MessageFlow(
                        outgoing_dict[flow_id][0],
                        incoming_dict[flow_id][0],
                        id=flow_id,
                        name=outgoing_dict[flow_id][3],
                        process=collaboration_id,
                    )
                elif flow_type.endswith("association"):
                    flow = BPMN.Association(
                        outgoing_dict[flow_id][0],
                        incoming_dict[flow_id][0],
                        id=flow_id,
                        name=outgoing_dict[flow_id][3],
                        process=outgoing_dict[flow_id][1],
                    )
                else:
                    flow = BPMN.SequenceFlow(
                        outgoing_dict[flow_id][0],
                        incoming_dict[flow_id][0],
                        id=flow_id,
                        name=outgoing_dict[flow_id][3],
                        process=outgoing_dict[flow_id][1],
                    )

                if flow is not None:
                    bpmn_graph.add_flow(flow)
                    layout.get(flow).del_waypoints()
                    for waypoint in flow_info[flow_id]:
                        layout.get(flow).add_waypoint(waypoint)

        for node_id in nodes_bounds:
            if node_id in nodes_dict:
                bounds = nodes_bounds[node_id]
                node = nodes_dict[node_id]
                layout.get(node).set_x(bounds["x"])
                layout.get(node).set_y(bounds["y"])
                layout.get(node).set_width(bounds["width"])
                layout.get(node).set_height(bounds["height"])
    return bpmn_graph


def import_xml_tree_from_root(root):
    """
    Imports a BPMN graph from (the root of) an XML tree

    Parameters
    -------------
    root
        Root of the tree

    Returns
    -------------
    bpmn_graph
        BPMN graph
    """
    bpmn_graph = BPMN()
    counts = Counts()
    incoming_dict = {}
    outgoing_dict = {}
    nodes_dict = {}
    nodes_bounds = {}
    flow_info = {}

    return parse_element(
        bpmn_graph,
        counts,
        root,
        [],
        incoming_dict,
        outgoing_dict,
        nodes_dict,
        nodes_bounds,
        flow_info,
    )


def apply(path, parameters=None):
    """
    Imports a BPMN diagram from a file

    Parameters
    -------------
    path
        Path to the file
    parameters
        Parameters of the algorithm

    Returns
    -------------
    bpmn_graph
        BPMN graph
    """
    if parameters is None:
        parameters = {}

    encoding = exec_utils.get_param_value(
        Parameters.ENCODING, parameters, None
    )

    from lxml import etree, objectify

    parser = etree.XMLParser(remove_comments=True, encoding=encoding)

    F = open(path, "rb")
    xml_tree = objectify.parse(F, parser=parser)
    F.close()

    return import_xml_tree_from_root(xml_tree.getroot())


def import_from_string(bpmn_string, parameters=None):
    """
    Imports a BPMN diagram from a string

    Parameters
    -------------
    path
        Path to the file
    parameters
        Parameters of the algorithm

    Returns
    -------------
    bpmn_graph
        BPMN graph
    """
    if parameters is None:
        parameters = {}

    encoding = exec_utils.get_param_value(
        Parameters.ENCODING, parameters, constants.DEFAULT_ENCODING
    )

    if type(bpmn_string) is str:
        bpmn_string = bpmn_string.encode(encoding)

    from lxml import etree, objectify

    parser = etree.XMLParser(remove_comments=True)
    root = objectify.fromstring(bpmn_string, parser=parser)

    return import_xml_tree_from_root(root)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util.colors import get_transitions_color


def get_alignments_decoration(
    net, im, fm, log=None, aligned_traces=None, parameters=None
):
    """
    Get a decoration for the Petri net based on alignments

    Parameters
    -------------
    net
        Petri net
    im
        Initial marking
    fm
        Final marking
    log
        Event log
    aligned_traces
        Aligned traces
    parameters
        Parameters of the algorithm

    Returns
    -------------
    decorations
        Decorations to use
    """
    if parameters is None:
        parameters = {}
    if aligned_traces is None and log is not None:
        from pm4py.algo.conformance.alignments.petri_net import (
            algorithm as alignments,
        )

        aligned_traces = alignments.apply(
            log, net, im, fm, parameters={"ret_tuple_as_trans_desc": True}
        )
    decorations = {}
    net_transitions = {}
    for trans in net.transitions:
        net_transitions[trans.name] = trans
    for align_trace0 in aligned_traces:
        align_trace = align_trace0["alignment"]
        for move in align_trace:
            move_trans_name = move[0][1]
            activity_trace_name = move[0][0]
            if move_trans_name in net_transitions:
                trans = net_transitions[move_trans_name]
                if trans not in decorations:
                    decorations[trans] = {
                        "count_fit": 0,
                        "count_move_on_model": 0,
                    }

                if activity_trace_name == ">>":
                    decorations[trans]["count_move_on_model"] = (
                        decorations[trans]["count_move_on_model"] + 1
                    )
                else:
                    decorations[trans]["count_fit"] = (
                        decorations[trans]["count_fit"] + 1
                    )

    for trans in decorations:
        if trans.label is not None:
            decorations[trans]["label"] = (
                trans.label
                + " ("
                + str(decorations[trans]["count_move_on_model"])
                + ","
                + str(decorations[trans]["count_fit"])
                + ")"
            )
            decorations[trans]["color"] = get_transitions_color(
                decorations[trans]["count_move_on_model"],
                decorations[trans]["count_fit"],
            )

    return decorations



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import importlib.util

from pm4py.algo.discovery.performance_spectrum.variants import (
    dataframe,
    log,
    dataframe_disconnected,
    log_disconnected,
)
from pm4py.util import exec_utils
from enum import Enum
from pm4py.util import constants, pandas_utils
from typing import Optional, Dict, Any, Union, List
from pm4py.objects.log.obj import EventLog, EventStream
import pandas as pd


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    ATTRIBUTE_KEY = constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY
    PARAMETER_SAMPLE_SIZE = "sample_size"


class Outputs(Enum):
    LIST_ACTIVITIES = "list_activities"
    POINTS = "points"


class Variants(Enum):
    DATAFRAME = dataframe
    LOG = log
    DATAFRAME_DISCONNECTED = dataframe_disconnected
    LOG_DISCONNECTED = log_disconnected


if importlib.util.find_spec("polars"):
    from pm4py.algo.discovery.performance_spectrum.variants import (
        lazyframe,
        lazyframe_disconnected,
    )

    Variants.LAZYFRAME = lazyframe
    Variants.LAZYFRAME_DISCONNECTED = lazyframe_disconnected


def is_polars_lazyframe(df: Any) -> bool:
    """Return True if the provided dataframe is a Polars LazyFrame."""
    df_type = str(type(df)).lower()
    return "polars" in df_type and "lazyframe" in df_type


def apply(
    log: Union[EventLog, EventStream, pd.DataFrame],
    list_activities: List[str],
    variant=None,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Dict[str, Any]:
    """
    Finds the performance spectrum provided a log/dataframe
    and a list of activities

    Parameters
    -------------
    log
        Event log/Dataframe
    list_activities
        List of activities interesting for the performance spectrum (at least two)
    variant
        Variant to be used (see Variants Enum)
    parameters
        Parameters of the algorithm, including:
            - Parameters.ACTIVITY_KEY
            - Parameters.TIMESTAMP_KEY

    Returns
    -------------
    ps
        Performance spectrum object (dictionary)
    """
    from pm4py.objects.conversion.log import converter as log_conversion

    if parameters is None:
        parameters = {}

    sample_size = exec_utils.get_param_value(
        Parameters.PARAMETER_SAMPLE_SIZE, parameters, 10000
    )

    if len(list_activities) < 2:
        raise Exception(
            "performance spectrum can be applied providing at least two activities!"
        )

    points = None

    if pandas_utils.check_is_pandas_dataframe(log):
        if variant is None:
            if is_polars_lazyframe(log):
                variant = Variants.LAZYFRAME
            else:
                variant = Variants.DATAFRAME

        points = exec_utils.get_variant(variant).apply(
            log, list_activities, sample_size, parameters
        )

    if points is None:
        if variant is None:
            variant = Variants.LOG

        points = exec_utils.get_variant(variant).apply(
            log_conversion.apply(
                log,
                variant=log_conversion.Variants.TO_EVENT_LOG,
                parameters=parameters,
            ),
            list_activities,
            sample_size,
            parameters,
        )

    ps = {
        Outputs.LIST_ACTIVITIES.value: list_activities,
        Outputs.POINTS.value: points,
    }

    return ps



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.visualization.decisiontree.variants import classic
from pm4py.visualization.common import gview
from pm4py.visualization.common import save as gsave
from enum import Enum
from pm4py.util import exec_utils
from pm4py.visualization.common.gview import serialize, serialize_dot
from sklearn.tree import DecisionTreeClassifier
from typing import Optional, Dict, Any, List
import graphviz


class Variants(Enum):
    CLASSIC = classic


DEFAULT_VARIANT = Variants.CLASSIC


def apply(
    clf: DecisionTreeClassifier,
    feature_names: List[str],
    classes: List[str],
    parameters: Optional[Dict[Any, Any]] = None,
    variant=DEFAULT_VARIANT,
) -> graphviz.Source:
    """
    Method to apply the visualization of the decision tree

    Parameters
    ------------
    clf
        Decision tree
    feature_names
        Names of the provided features
    classes
        Names of the target classes
    parameters
        Possible parameters of the algorithm, including:
            Parameters.FORMAT -> Image format (pdf, svg, png ...)
    variant
        Variant of the algorithm:
            - Variants.CLASSIC

    Returns
    ------------
    gviz
        GraphViz object
    """
    return exec_utils.get_variant(variant).apply(
        clf, feature_names, classes, parameters=parameters
    )


def save(gviz: graphviz.Source, output_file_path: str, parameters=None):
    """
    Save the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    output_file_path
        Path where the GraphViz output should be saved
    """
    gsave.save(gviz, output_file_path, parameters=parameters)


def view(gviz: graphviz.Source, parameters=None):
    """
    View the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    """
    return gview.view(gviz, parameters=parameters)


def matplotlib_view(gviz: graphviz.Source, parameters=None):
    """
    Views the diagram using Matplotlib

    Parameters
    ---------------
    gviz
        Graphviz
    """

    return gview.matplotlib_view(gviz, parameters=parameters)



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import Optional, Dict, Any, List, Set

import numpy as np
import pandas as pd

from pm4py.objects.log.util.dataframe_utils import Parameters
from pm4py.util import constants, exec_utils, pandas_utils
from pm4py.util import xes_constants


def _sanitize_feature_name(
    prefix: str, value: Any, used_names: Optional[Set[str]] = None
) -> str:
    """
    Build a deterministic feature column name and ensure uniqueness when requested.
    """
    sanitized = (
        str(value)
        .encode("ascii", errors="ignore")
        .decode("ascii")
    )
    if sanitized == "":
        sanitized = "value"
    base_name = f"{prefix}_{sanitized}"

    if used_names is None:
        return base_name

    candidate = base_name
    suffix = 1
    while candidate in used_names:
        candidate = f"{base_name}__{suffix}"
        suffix += 1
    used_names.add(candidate)
    return candidate


def automatic_feature_selection_df(df, parameters=None):
    """
    Performs an automatic feature selection on dataframes,
    keeping the features useful for ML purposes

    Parameters
    ---------------
    df
        Dataframe
    parameters
        Parameters of the algorithm

    Returns
    ---------------
    featured_df
        Dataframe with only the features that have been selected
    """
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )

    mandatory_attributes = exec_utils.get_param_value(
        Parameters.MANDATORY_ATTRIBUTES,
        parameters,
        set(df.columns).intersection(
            {case_id_key, activity_key, timestamp_key}
        ),
    )

    min_different_occ_str_attr = exec_utils.get_param_value(
        Parameters.MIN_DIFFERENT_OCC_STR_ATTR, parameters, 5
    )
    max_different_occ_str_attr = exec_utils.get_param_value(
        Parameters.MAX_DIFFERENT_OCC_STR_ATTR, parameters, 50
    )
    consider_all_attributes = exec_utils.get_param_value(
        Parameters.CONSIDER_ALL_ATTRIBUTES, parameters, True
    )

    cols_dtypes = {x: str(df[x].dtype) for x in df.columns}
    other_attributes_to_retain = set()

    no_all_cases = df[case_id_key].nunique()
    for x, y in cols_dtypes.items():
        attr_df = df.dropna(subset=[x])
        this_cases = attr_df[case_id_key].nunique()
        attr_in_all_cases = this_cases == no_all_cases

        # in any case, keep attributes that appears at least once per case
        if attr_in_all_cases or consider_all_attributes:
            if "float" in y or "int" in y:
                # (as in the classic log version) retain always float/int attributes
                other_attributes_to_retain.add(x)
            elif "obj" in y or "str" in y:
                # (as in the classic log version) keep string attributes if they have enough variability, but not too much
                # (that would be hard to explain)
                unique_val_count = df[x].nunique()
                if (
                    min_different_occ_str_attr
                    <= unique_val_count
                    <= max_different_occ_str_attr
                ):
                    other_attributes_to_retain.add(x)
            else:
                # not consider the attribute after this feature selection if it
                # has other types (for example, date)
                pass

    attributes_to_retain = mandatory_attributes.union(
        other_attributes_to_retain
    )

    return df[list(attributes_to_retain)]


def select_number_column(
    df: pd.DataFrame,
    fea_df: pd.DataFrame,
    col: str,
    case_id_key=constants.CASE_CONCEPT_NAME,
) -> pd.DataFrame:
    """
    Extract a column for the features dataframe for the given numeric attribute

    Parameters
    --------------
    df
        Dataframe
    fea_df
        Feature dataframe
    col
        Numeric column
    case_id_key
        Case ID key

    Returns
    --------------
    fea_df
        Feature dataframe (desidered output)
    """
    # More efficient: drop duplicates keeping last instead of groupby
    df_subset = df[[case_id_key, col]].dropna(subset=[col])
    df_last = df_subset.drop_duplicates(subset=[case_id_key], keep="last")

    fea_df = fea_df.merge(
        df_last, on=[case_id_key], how="left", suffixes=("", "_y")
    )
    fea_df[col] = fea_df[col].astype(np.float32)
    return fea_df


def select_string_column(
    df: pd.DataFrame,
    fea_df: pd.DataFrame,
    col: str,
    case_id_key=constants.CASE_CONCEPT_NAME,
    count_occurrences=False,
) -> pd.DataFrame:
    """
    Extract N columns (for N different attribute values; hotencoding) for the features dataframe for the given string attribute

    Parameters
    --------------
    df
        Dataframe
    fea_df
        Feature dataframe
    col
        String column
    case_id_key
        Case ID key
    count_occurrences
        If True, count the number of occurrences of the attribute value in each case.
        If False (default), use binary encoding (1 if present, 0 if not present)

    Returns
    --------------
    fea_df
        Feature dataframe (desidered output)
    """
    # Filter out None values once
    df_filtered = df[[case_id_key, col]].dropna(subset=[col])
    existing_cols = set(fea_df.columns)

    if count_occurrences:
        # Use crosstab for efficient counting
        crosstab = pd.crosstab(df_filtered[case_id_key], df_filtered[col])
        # Rename columns efficiently while avoiding duplicates
        rename_map = {
            original: _sanitize_feature_name(col, original, existing_cols)
            for original in crosstab.columns
        }
        crosstab = crosstab.rename(columns=rename_map)
        # Merge once with all columns
        fea_df = fea_df.merge(
            crosstab, left_on=case_id_key, right_index=True, how="left"
        )
        # Fill NaN and convert to float32 for all new columns at once
        new_cols = list(rename_map.values())
        if new_cols:
            fea_df[new_cols] = fea_df[new_cols].astype(np.float32)
    else:
        # Use pivot_table for binary encoding - much faster than loop
        # Create a dummy column for aggregation
        df_filtered = df_filtered.copy()
        df_filtered["_dummy"] = 1
        cases_with_values = df_filtered[case_id_key].unique()

        # Get unique values
        unique_vals = pandas_utils.format_unique(df_filtered[col].unique())
        unique_vals = [v for v in unique_vals if v is not None]

        if len(unique_vals) > 0:
            # Create pivot table
            pivot = df_filtered.pivot_table(
                index=case_id_key,
                columns=col,
                values="_dummy",
                aggfunc="max",
            )

            # Rename columns (ensuring uniqueness when sanitization collides)
            rename_map = {
                original: _sanitize_feature_name(col, original, existing_cols)
                for original in pivot.columns
            }
            pivot = pivot.rename(columns=rename_map)

            # Merge once with all columns
            fea_df = fea_df.merge(
                pivot, left_on=case_id_key, right_index=True, how="left"
            )
            # Fill NaN only for cases having at least one value and convert to float32
            new_cols = list(rename_map.values())
            if new_cols:
                mask = (
                    fea_df[case_id_key].isin(cases_with_values)
                    if case_id_key in fea_df.columns
                    else None
                )
                for col_name in new_cols:
                    if col_name in fea_df.columns:
                        if mask is not None:
                            fea_df.loc[mask, col_name] = (
                                fea_df.loc[mask, col_name].fillna(0)
                            )
                        fea_df[col_name] = fea_df[col_name].astype(np.float32)

    return fea_df


def get_features_df(
    df: pd.DataFrame,
    list_columns: List[str],
    parameters: Optional[Dict[Any, Any]] = None,
) -> pd.DataFrame:
    """
    Given a dataframe and a list of columns, performs an automatic feature extraction

    Parameters
    ---------------
    df
        Dataframe
    list_column
        List of column to consider in the feature extraction
    parameters
        Parameters of the algorithm, including:
        - Parameters.CASE_ID_KEY: the case ID
        - Parameters.COUNT_OCCURRENCES: if True, count occurrences of string attributes instead of binary encoding

    Returns
    ---------------
    fea_df
        Feature dataframe (desidered output)
    """
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    add_case_identifier_column = exec_utils.get_param_value(
        Parameters.ADD_CASE_IDENTIFIER_COLUMN, parameters, False
    )
    count_occurrences = exec_utils.get_param_value(
        Parameters.COUNT_OCCURRENCES, parameters, False
    )

    # Start with unique case IDs
    unique_cases = sorted(pandas_utils.format_unique(df[case_id_key].unique()))
    fea_df = pandas_utils.instantiate_dataframe({case_id_key: unique_cases})

    # Separate columns by type for batch processing
    string_columns = []
    numeric_columns = []

    for col in list_columns:
        dtype_str = str(df[col].dtype)
        if "obj" in dtype_str or "str" in dtype_str:
            string_columns.append(col)
        elif "float" in dtype_str or "int" in dtype_str:
            numeric_columns.append(col)

    # Process numeric columns (can be done more efficiently in batch)
    if numeric_columns:
        # Process all numeric columns at once
        df_numeric = df[[case_id_key] + numeric_columns].copy()
        # Drop duplicates for all numeric columns at once, keeping last
        df_numeric = df_numeric.groupby(case_id_key).last().reset_index()
        # Merge once for all numeric columns
        fea_df = fea_df.merge(df_numeric, on=case_id_key, how="left")
        # Convert all numeric columns to float32 at once
        fea_df[numeric_columns] = fea_df[numeric_columns].astype(np.float32)

    # Process string columns one by one (still needed due to hot encoding)
    for col in string_columns:
        fea_df = select_string_column(
            df, fea_df, col, case_id_key=case_id_key, count_occurrences=count_occurrences
        )

    # Sort and optionally remove case ID
    fea_df = fea_df.sort_values(case_id_key)
    if not add_case_identifier_column:
        del fea_df[case_id_key]

    return fea_df


def automatic_feature_extraction_df(
    df: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None
) -> pd.DataFrame:
    """
    Performs an automatic feature extraction given a dataframe

    Parameters
    --------------
    df
        Dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.CASE_ID_KEY: the case ID
        - Parameters.MIN_DIFFERENT_OCC_STR_ATTR
        - Parameters.MAX_DIFFERENT_OCC_STR_ATTR

    Returns
    --------------
    fea_df
        Dataframe with the features
    """
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )

    fea_sel_df = automatic_feature_selection_df(df, parameters=parameters)
    columns = set(fea_sel_df.columns)

    if case_id_key in columns:
        columns.remove(case_id_key)

    if timestamp_key in columns:
        columns.remove(timestamp_key)

    return get_features_df(fea_sel_df, list(columns), parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.process_tree.exporter.variants import ptml
from pm4py.util import exec_utils
from enum import Enum


class Variants(Enum):
    PTML = ptml


DEFAULT_VARIANT = Variants.PTML


def apply(tree, output_path, variant=DEFAULT_VARIANT, parameters=None):
    """
    Exports the process tree to a file

    Parameters
    ----------------
    tree
        Process tree
    output_path
        Output path
    variant
        Variant of the algorithm:
            - Variants.PTML
    parameters
        Parameters
    """
    return exec_utils.get_variant(variant).apply(
        tree, output_path, parameters=parameters
    )


def serialize(tree, variant=DEFAULT_VARIANT, parameters=None):
    """
    Serializes the process tree into a binary string

    Parameters
    ----------------
    tree
        Process tree
    variant
        Variant of the algorithm:
            - Variants.PTML
    parameters
        Parameters

    Returns
    ---------------
    serialization
        Serialized string
    """
    return exec_utils.get_variant(variant).export_tree_as_string(
        tree, parameters=parameters
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import Dict, Tuple


def apply(dfg: Dict[Tuple[str, str], int]) -> Dict[Tuple[str, str], float]:
    """
    Computes a causal graph based on a directly follows graph according to the heuristics miner

    Parameters
    ----------
    dfg: :class:`dict` directly follows relation, should be a dict of the form (activity,activity) -> num of occ.

    Returns
    -------
    :return: dictionary containing all causal relations as keys (with value inbetween -1 and 1 indicating that
    how strong it holds)
    """
    causal_heur = {}
    for f, t in dfg:
        if (f, t) not in causal_heur:
            rev = dfg[(t, f)] if (t, f) in dfg else 0
            causal_heur[(f, t)] = float(
                (dfg[(f, t)] - rev) / (dfg[(f, t)] + rev + 1)
            )
            causal_heur[(t, f)] = -1 * causal_heur[(f, t)]
    return causal_heur



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog, EventStream
import pandas as pd
from enum import Enum
from pm4py.util import exec_utils
from pm4py.algo.label_splitting.variants import contextual


class Variants(Enum):
    CONTEXTUAL = contextual


def apply(
    log: Union[EventLog, EventStream, pd.DataFrame],
    variant=Variants.CONTEXTUAL,
    parameters: Optional[Dict[Any, Any]] = None,
) -> pd.DataFrame:
    """
    Applies a technique of label-splitting, to distinguish between different meanings of the same
    activity. The result is a Pandas dataframe where the label-splitting has been applied.

    Minimum Viable Example:

        import pm4py
        from pm4py.algo.label_splitting import algorithm as label_splitter

        log = pm4py.read_xes("tests/input_data/receipt.xes")
        log2 = label_splitter.apply(log)


    Parameters
    ---------------
    log
        Event log
    parameters
        Variant-specific parameters

    Returns
    ---------------
    dataframe
        Pandas dataframe with the re-labeling
    """
    return exec_utils.get_variant(variant).apply(log, parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import pandas as pd
import importlib.util

from pm4py.util import constants, xes_constants
import numpy as np


def get_default_dataframe_environment():
    if importlib.util.find_spec("cudf"):
        # import cudf; return cudf
        try:
            import cudf.pandas

            cudf.pandas.install()
        except BaseException:
            pass
    import pandas as pd

    return pd


DATAFRAME = get_default_dataframe_environment()


def to_dict_records(df):
    """
    Pandas dataframe to dictionary (records method)

    Parameters
    ---------------
    df
        Dataframe

    Returns
    --------------
    list_dictio
        List containing a dictionary for each row
    """
    if is_polars_lazyframe(df):
        return df.collect().to_dicts()

    return df.to_dict("records")


def to_dict_index(df):
    """
    Pandas dataframe to dictionary (index method)

    Parameters
    ---------------
    df
        Dataframe

    Returns
    --------------
    dict
        dict like {index -> {column -> value}}
    """
    if is_polars_lazyframe(df):
        collected_df = df.collect()
        return {
            idx: row
            for idx, row in enumerate(collected_df.iter_rows(named=True))
        }

    return df.to_dict("index")


def insert_index(
    df,
    column_name=constants.DEFAULT_INDEX_KEY,
    copy_dataframe=True,
    reset_index=True,
):
    """
    Inserts the dataframe index in the specified column

    Parameters
    --------------
    df
        Dataframe
    column_name
        Name of the column that should host the index
    copy_dataframe
        Establishes if the original dataframe should be copied before inserting the column

    Returns
    --------------
    df
        Dataframe with index
    """
    if is_polars_lazyframe(df):
        lf = df
        existing_columns = set(_lazyframe_column_names(lf))
        if column_name in existing_columns:
            lf = lf.drop(column_name)
        return lf.with_row_count(name=column_name)

    if copy_dataframe:
        df = df.copy()

    if reset_index:
        df = df.reset_index(drop=True)

    df[column_name] = df.index
    return df


def insert_case_index(
    df,
    column_name=constants.DEFAULT_CASE_INDEX_KEY,
    case_id=constants.CASE_CONCEPT_NAME,
    copy_dataframe=True,
):
    """
    Inserts the case number in the dataframe

    Parameters
    ---------------
    df
        Dataframe
    column_name
        Name of the column that should host the case index
    case_id
        Case identifier
    copy_dataframe
        Establishes if the original dataframe should be copied before inserting the column

    Returns
    ---------------
    df
        Dataframe with case index
    """
    if is_polars_lazyframe(df):
        import polars as pl  # type: ignore[import-untyped]

        lf = df
        existing_columns = set(_lazyframe_column_names(lf))
        if column_name in existing_columns:
            lf = lf.drop(column_name)
        case_map = (
            lf.select(pl.col(case_id))
            .unique(maintain_order=True)
            .with_row_count(name=column_name)
        )
        return lf.join(case_map, on=case_id, how="left")

    if copy_dataframe:
        df = df.copy()

    df[column_name] = df.groupby(case_id).ngroup()
    return df


def insert_ev_in_tr_index(
    df: pd.DataFrame,
    case_id: str = constants.CASE_CONCEPT_NAME,
    column_name: str = constants.DEFAULT_INDEX_IN_TRACE_KEY,
    copy_dataframe=True,
) -> pd.DataFrame:
    """
    Inserts a column that specify the index of the event inside the case

    Parameters
    ---------------
    df
        Dataframe
    case_id
        Column that hosts the case identifier
    column_name
        Name of the column that should host the index

    Returns
    --------------
    df
        Dataframe with index
    """
    if is_polars_lazyframe(df):
        import polars as pl  # type: ignore[import-untyped]

        lf = df
        existing_columns = set(_lazyframe_column_names(lf))
        if column_name in existing_columns:
            lf = lf.drop(column_name)
        return lf.with_columns(
            (pl.col(case_id).cum_count().over(case_id) - 1).alias(column_name)
        )

    if copy_dataframe:
        df = df.copy()

    df_trace_idx = df.groupby(case_id).cumcount()
    df[column_name] = df_trace_idx
    return df


def format_unique(values):
    try:
        values = values.to_numpy()
    except BaseException:
        pass

    try:
        values = values.tolist()
    except AttributeError:
        try:
            values = values.to_list()
        except AttributeError:
            values = list(values)

    return values


def insert_feature_activity_position_in_trace(
    df: pd.DataFrame,
    case_id: str = constants.CASE_CONCEPT_NAME,
    activity_key: str = xes_constants.DEFAULT_NAME_KEY,
    prefix="@@position_",
):
    """
    Inserts additional columns @@position_ACT1, @@position_ACT2 ...
    which are populated for every event having activity ACT1, ACT2 respectively,
    with the index of the event inside its case.

    Parameters
    ------------------
    df
        Pandas dataframe
    case_id
        Case idntifier
    activity_key
        Activity
    prefix
        Prefix of the "activity position in trace" feature (default: @@position_)

    Returns
    ------------------
    df
        Pandas dataframe
    """
    if is_polars_lazyframe(df):
        import polars as pl  # type: ignore[import-untyped]

        lf = insert_ev_in_tr_index(
            df,
            case_id=case_id,
            column_name=constants.DEFAULT_INDEX_IN_TRACE_KEY,
            copy_dataframe=False,
        )
        activities = (
            lf.select(pl.col(activity_key))
            .unique()
            .collect()
            .get_column(activity_key)
            .to_list()
        )
        existing_columns = set(_lazyframe_column_names(lf))
        for act in activities:
            column_name = prefix + str(act)
            if column_name in existing_columns:
                lf = lf.drop(column_name)
                existing_columns.discard(column_name)
            lf = lf.with_columns(
                pl.when(pl.col(activity_key) == act)
                .then(pl.col(constants.DEFAULT_INDEX_IN_TRACE_KEY))
                .otherwise(pl.lit(None))
                .alias(column_name)
            )
            existing_columns.add(column_name)
        return lf

    df = insert_ev_in_tr_index(df, case_id=case_id)
    activities = format_unique(df[activity_key].unique())
    for act in activities:
        df[prefix + act] = df[activity_key].apply(
            lambda x: np.nan if x == act else -1
        )
        df[prefix + act] = df[prefix + act].fillna(
            df[constants.DEFAULT_INDEX_IN_TRACE_KEY]
        )
        df[prefix + act] = df[prefix + act].replace(-1, np.nan)
    return df


def insert_case_arrival_finish_rate(
    log: pd.DataFrame,
    case_id_column=constants.CASE_CONCEPT_NAME,
    timestamp_column=xes_constants.DEFAULT_TIMESTAMP_KEY,
    start_timestamp_column=None,
    arrival_rate_column="@@arrival_rate",
    finish_rate_column="@@finish_rate",
) -> pd.DataFrame:
    """
    Inserts the arrival/finish rate in the dataframe.

    Parameters
    -----------------
    log
        Pandas dataframe

    Returns
    -----------------
    log
        Pandas dataframe enriched by arrival and finish rate
    """
    if is_polars_lazyframe(log):
        import polars as pl  # type: ignore[import-untyped]

        if start_timestamp_column is None:
            start_timestamp_column = timestamp_column

        lf = log
        existing_columns = set(_lazyframe_column_names(lf))
        if arrival_rate_column in existing_columns:
            lf = lf.drop(arrival_rate_column)
            existing_columns.discard(arrival_rate_column)
        if finish_rate_column in existing_columns:
            lf = lf.drop(finish_rate_column)

        arrival = (
            lf.select(
                pl.col(case_id_column),
                pl.col(start_timestamp_column).alias("__start_ts"),
            )
            .group_by(case_id_column)
            .agg(pl.col("__start_ts").min().alias("__start_ts"))
            .with_columns(
                pl.col("__start_ts")
                .dt.timestamp()
                .alias("__arrival_microseconds")
            )
            .sort(["__arrival_microseconds", case_id_column])
            .with_columns(
                pl.col("__arrival_microseconds")
                .diff()
                .fill_null(0)
                .truediv(1_000_000)
                .alias(arrival_rate_column)
            )
            .select(case_id_column, arrival_rate_column)
        )

        finish = (
            lf.select(
                pl.col(case_id_column),
                pl.col(timestamp_column).alias("__finish_ts"),
            )
            .group_by(case_id_column)
            .agg(pl.col("__finish_ts").max().alias("__finish_ts"))
            .with_columns(
                pl.col("__finish_ts")
                .dt.timestamp()
                .alias("__finish_microseconds")
            )
            .sort(["__finish_microseconds", case_id_column])
            .with_columns(
                pl.col("__finish_microseconds")
                .diff()
                .fill_null(0)
                .truediv(1_000_000)
                .alias(finish_rate_column)
            )
            .select(case_id_column, finish_rate_column)
        )

        lf = lf.join(arrival, on=case_id_column, how="left")
        lf = lf.join(finish, on=case_id_column, how="left")
        return lf

    if start_timestamp_column is None:
        start_timestamp_column = timestamp_column

    case_arrival = (
        log.groupby(case_id_column)[start_timestamp_column]
        .agg("min")
        .to_dict()
    )
    case_arrival = [[x, y.timestamp()] for x, y in case_arrival.items()]
    case_arrival.sort(key=lambda x: (x[1], x[0]))

    case_finish = (
        log.groupby(case_id_column)[timestamp_column].agg("max").to_dict()
    )
    case_finish = [[x, y.timestamp()] for x, y in case_finish.items()]
    case_finish.sort(key=lambda x: (x[1], x[0]))

    i = len(case_arrival) - 1
    while i > 0:
        case_arrival[i][1] = case_arrival[i][1] - case_arrival[i - 1][1]
        i = i - 1
    case_arrival[0][1] = 0
    case_arrival = {x[0]: x[1] for x in case_arrival}

    i = len(case_finish) - 1
    while i > 0:
        case_finish[i][1] = case_finish[i][1] - case_finish[i - 1][1]
        i = i - 1
    case_finish[0][1] = 0
    case_finish = {x[0]: x[1] for x in case_finish}

    log[arrival_rate_column] = log[case_id_column].map(case_arrival)
    log[finish_rate_column] = log[case_id_column].map(case_finish)

    return log


def insert_case_service_waiting_time(
    log: pd.DataFrame,
    case_id_column=constants.CASE_CONCEPT_NAME,
    timestamp_column=xes_constants.DEFAULT_TIMESTAMP_KEY,
    start_timestamp_column=None,
    diff_start_end_column="@@diff_start_end",
    service_time_column="@@service_time",
    sojourn_time_column="@@sojourn_time",
    waiting_time_column="@@waiting_time",
) -> pd.DataFrame:
    """
    Inserts the service/waiting/sojourn time in the dataframe.

    Parameters
    ----------------
    log
        Pandas dataframe
    parameters
        Parameters of the method

    Returns
    ----------------
    log
        Pandas dataframe with service, waiting and sojourn time
    """
    if is_polars_lazyframe(log):
        import polars as pl  # type: ignore[import-untyped]

        if start_timestamp_column is None:
            start_timestamp_column = timestamp_column

        lf = log
        existing_columns = set(_lazyframe_column_names(lf))
        for col_name in (
            diff_start_end_column,
            service_time_column,
            sojourn_time_column,
            waiting_time_column,
        ):
            if col_name in existing_columns:
                lf = lf.drop(col_name)
                existing_columns.discard(col_name)

        lf = lf.with_columns(
            (
                pl.col(timestamp_column).dt.timestamp()
                - pl.col(start_timestamp_column).dt.timestamp()
            )
            .truediv(1_000_000)
            .alias(diff_start_end_column)
        )

        service = (
            lf.group_by(case_id_column)
            .agg(
                pl.col(diff_start_end_column)
                .sum()
                .alias(service_time_column)
            )
        )

        sojourn = (
            lf.group_by(case_id_column)
            .agg(
                pl.col(start_timestamp_column)
                .dt.timestamp()
                .min()
                .alias("__case_start_ts"),
                pl.col(timestamp_column)
                .dt.timestamp()
                .max()
                .alias("__case_end_ts"),
            )
            .with_columns(
                (
                    pl.col("__case_end_ts") - pl.col("__case_start_ts")
                )
                .truediv(1_000_000)
                .alias(sojourn_time_column)
            )
            .select(case_id_column, sojourn_time_column)
        )

        lf = lf.join(service, on=case_id_column, how="left")
        lf = lf.join(sojourn, on=case_id_column, how="left")
        lf = lf.with_columns(
            (
                pl.col(sojourn_time_column) - pl.col(service_time_column)
            ).alias(waiting_time_column)
        )
        return lf

    if start_timestamp_column is None:
        start_timestamp_column = timestamp_column

    log[diff_start_end_column] = get_total_seconds(
        log[timestamp_column] - log[start_timestamp_column]
    )
    service_times = (
        log.groupby(case_id_column)[diff_start_end_column].sum().to_dict()
    )
    log[service_time_column] = log[case_id_column].map(service_times)

    start_timestamps = (
        log.groupby(case_id_column)[start_timestamp_column]
        .agg("min")
        .to_dict()
    )
    complete_timestamps = (
        log.groupby(case_id_column)[timestamp_column].agg("max").to_dict()
    )
    sojourn_time_cases = {
        x: complete_timestamps[x].timestamp() - start_timestamps[x].timestamp()
        for x in start_timestamps
    }

    log[sojourn_time_column] = log[case_id_column].map(sojourn_time_cases)
    log[waiting_time_column] = (
        log[sojourn_time_column] - log[service_time_column]
    )

    return log


def check_is_pandas_dataframe(log):
    """
    Checks if a log object is a dataframe

    Parameters
    -------------
    log
        Log object

    Returns
    -------------
    boolean
        Is dataframe?
    """
    log_type = str(type(log)).lower()
    return "dataframe" in log_type or "lazyframe" in log_type


def is_polars_lazyframe(df):
    """Return True if the provided dataframe is a Polars LazyFrame."""
    df_type = str(type(df)).lower()
    return "polars" in df_type and "lazyframe" in df_type


def _lazyframe_schema(df):
    """Return the Polars schema without triggering expensive resolution via .columns/.schema."""
    return df.collect_schema()


def _lazyframe_column_names(df):
    """Return column names for a Polars LazyFrame without collecting the frame."""
    return _lazyframe_schema(df).names()


def instantiate_dataframe(*args, **kwargs):
    return DATAFRAME.DataFrame(*args, **kwargs)


def instantiate_dataframe_from_dict(*args, **kwargs):
    return DATAFRAME.DataFrame.from_dict(*args, **kwargs)


def instantiate_dataframe_from_records(*args, **kwargs):
    return DATAFRAME.DataFrame.from_records(*args, **kwargs)


def get_grouper(*args, **kwargs):
    return DATAFRAME.Grouper(*args, **kwargs)


def get_total_seconds(difference):
    return (
        86400 * difference.dt.days
        + difference.dt.seconds
        + 10**-6 * difference.dt.microseconds
        + 10**-9 * difference.dt.nanoseconds
    )


def convert_to_seconds(dt_column):
    try:
        # Pandas
        return dt_column.values.astype(np.int64) / 10**9
    except BaseException:
        # CUDF
        return [x / 10**9 for x in dt_column.to_numpy().tolist()]


def dataframe_column_string_to_datetime(*args, **kwargs):
    if (
        importlib.util.find_spec("cudf")
        or constants.TEST_CUDF_DATAFRAMES_ENVIRONMENT
    ):
        pass
        """if DATAFRAME == pd:
            format = kwargs["format"] if "format" in kwargs else None
            if format not in [None, 'mixed', 'ISO8601']:
                kwargs["exact"] = False"""

    return DATAFRAME.to_datetime(*args, **kwargs)


def read_csv(*args, **kwargs):
    if (
        importlib.util.find_spec("cudf")
        or constants.TEST_CUDF_DATAFRAMES_ENVIRONMENT
    ):
        if kwargs and "encoding" in kwargs:
            del kwargs["encoding"]

    return DATAFRAME.read_csv(*args, **kwargs)


def concat(*args, **kwargs):
    return DATAFRAME.concat(*args, **kwargs)


def merge(*args, **kwargs):
    return DATAFRAME.merge(*args, **kwargs)


def check_pandas_dataframe_columns(
    df,
    activity_key=None,
    case_id_key=None,
    timestamp_key=None,
    start_timestamp_key=None,
):
    """
    Checks if the dataframe contains all the required columns.
    If not, raise an exception

    Parameters
    --------------
    df
        Pandas dataframe
    """
    if is_polars_lazyframe(df):
        import polars as pl  # type: ignore[import-untyped]

        schema = _lazyframe_schema(df)
        columns = list(schema.names())
        column_set = set(columns)
        if len(columns) < 3:
            raise Exception(
                "the dataframe should (at least) contain a column for the case identifier, a column for the activity and a column for the timestamp."
            )

        str_columns = {
            col
            for col, dtype in schema.items()
            if any(
                token in str(dtype).lower()
                for token in ("str", "utf8", "categorical", "enum", "object")
            )
        }
        timest_columns = {
            col
            for col, dtype in schema.items()
            if any(token in str(dtype).lower() for token in ("date", "time", "datetime"))
        }

        if len(str_columns) < 2:
            raise Exception(
                "the dataframe should (at least) contain a column of type string for the case identifier and a column of type string for the activity."
            )

        if len(timest_columns) < 1:
            raise Exception("the dataframe should (at least) contain a column of type date")

        def raise_if_missing(column_name):
            raise Exception(
                "the specified {} column is not contained in the dataframe. Available columns: {}".format(
                    column_name,
                    sorted(list(columns)),
                )
            )

        def ensure_no_null(col, label):
            has_null_df = df.select(
                pl.col(col).is_null().any().alias("__has_null")
            ).collect()
            if bool(has_null_df[0, "__has_null"]):
                raise Exception(
                    "the {} column should not contain any empty value.".format(label)
                )

        if case_id_key is not None:
            if case_id_key not in column_set:
                raise_if_missing("case ID")
            if case_id_key not in str_columns:
                raise Exception("the case ID column should be of type string.")
            ensure_no_null(case_id_key, "case ID")

        if activity_key is not None:
            if activity_key not in column_set:
                raise_if_missing("activity")
            if activity_key not in str_columns:
                raise Exception("the activity column should be of type string.")
            ensure_no_null(activity_key, "activity")

        if timestamp_key is not None:
            if timestamp_key not in column_set:
                raise_if_missing("timestamp")
            if timestamp_key not in timest_columns:
                raise Exception(
                    "the timestamp column should be of type datetime. Use the function pandas.to_datetime"
                )
            ensure_no_null(timestamp_key, "timestamp")

        if start_timestamp_key is not None:
            if start_timestamp_key not in column_set:
                raise_if_missing("start timestamp")
            if start_timestamp_key not in timest_columns:
                raise Exception(
                    "the start timestamp column should be of type datetime. Use the function pandas.to_datetime"
                )
            ensure_no_null(start_timestamp_key, "start timestamp")

        return

    if len(df.columns) < 3:
        raise Exception(
            "the dataframe should (at least) contain a column for the case identifier, a column for the activity and a column for the timestamp."
        )

    str_columns = {
        x
        for x in df.columns
        if "str" in str(df[x].dtype).lower()
        or "obj" in str(df[x].dtype).lower()
    }
    timest_columns = {
        x
        for x in df.columns
        if "date" in str(df[x].dtype).lower()
        or "time" in str(df[x].dtype).lower()
    }

    if len(str_columns) < 2:
        raise Exception(
            "the dataframe should (at least) contain a column of type string for the case identifier and a column of type string for the activity."
        )

    if len(timest_columns) < 1:
        raise Exception(
            "the dataframe should (at least) contain a column of type date"
        )

    if case_id_key is not None:
        if case_id_key not in df.columns:
            raise Exception(
                "the specified case ID column is not contained in the dataframe. Available columns: "
                + str(sorted(list(df.columns)))
            )

        if case_id_key not in str_columns:
            raise Exception("the case ID column should be of type string.")

        if df[case_id_key].isnull().values.any():
            raise Exception(
                "the case ID column should not contain any empty value."
            )

    if activity_key is not None:
        if activity_key not in df.columns:
            raise Exception(
                "the specified activity column is not contained in the dataframe. Available columns: "
                + str(sorted(list(df.columns)))
            )

        if activity_key not in str_columns:
            raise Exception("the activity column should be of type string.")

        if df[activity_key].isnull().values.any():
            raise Exception(
                "the activity column should not contain any empty value."
            )

    if timestamp_key is not None:
        if timestamp_key not in df.columns:
            raise Exception(
                "the specified timestamp column is not contained in the dataframe. Available columns: "
                + str(sorted(list(df.columns)))
            )

        if timestamp_key not in timest_columns:
            raise Exception(
                "the timestamp column should be of time datetime. Use the function pandas.to_datetime"
            )

        if df[timestamp_key].isnull().values.any():
            raise Exception(
                "the timestamp column should not contain any empty value."
            )

    if start_timestamp_key is not None:
        if start_timestamp_key not in df.columns:
            raise Exception(
                "the specified start timestamp column is not contained in the dataframe. Available columns: "
                + str(sorted(list(df.columns)))
            )

        if start_timestamp_key not in timest_columns:
            raise Exception(
                "the start timestamp column should be of time datetime. Use the function pandas.to_datetime"
            )

        if df[start_timestamp_key].isnull().values.any():
            raise Exception(
                "the start timestamp column should not contain any empty value."
            )


def get_traces(log, case_id_key, activity_key):
    if is_polars_lazyframe(log):
        import polars as pl  # type: ignore[import-untyped]

        collected = (
            log.select(pl.col(case_id_key), pl.col(activity_key))
            .group_by(case_id_key, maintain_order=True)
            .agg(pl.col(activity_key).alias("__acts"))
            .collect()
        )
        traces = [tuple(seq) for seq in collected["__acts"].to_list()]
    else:
        # Pandas dataframe management
        traces = [
            tuple(x)
            for x in log.groupby(case_id_key)[activity_key]
            .agg(list)
            .to_dict()
            .values()
        ]
    return traces


def get_attribute_values_count(log, attribute):
    if is_polars_lazyframe(log):
        import importlib.util

        if importlib.util.find_spec("polars") is None:
            raise RuntimeError(
                "Polars LazyFrame provided but 'polars' package is not installed."
            )

        import polars as pl  # type: ignore[import-untyped]

        available_columns = set(_lazyframe_column_names(log))
        if attribute not in available_columns:
            raise Exception(
                f"Column '{attribute}' is not present in the provided Polars LazyFrame."
            )

        result = (
            log.group_by(attribute)
            .agg(pl.len().alias("__pm4py_count__"))
            .collect()
        )

        return {
            row[attribute]: row["__pm4py_count__"]
            for row in result.iter_rows(named=True)
        }
    else:
        return log[attribute].value_counts().to_dict()


def df_row_count(log):
    if is_polars_lazyframe(log):
        return len(log.collect())
    else:
        return len(log)


def get_pivot_timestamp_distribution(
        dataframe: pd.DataFrame,
        frequency_alias="M",
        case_id_col="case:concept:name",
        timestamp_col="time:timestamp"
) -> pd.DataFrame:
    """
    Creates a pivot table showing the distribution of timestamp occurrences for each case,
    grouped by a specified time frequency.

    Parameters
    ----------
    dataframe : pd.DataFrame
        The input event log as a pandas DataFrame, with at least case and timestamp columns.
    frequency_alias : str, default 'M'
        The frequency at which to group (bin) the timestamps. This value should be a valid
        pandas frequency alias, such as:
            - 'A' or 'Y'   : Yearly (e.g., 2024)
            - 'Q'          : Quarterly (e.g., 2024Q2)
            - 'M'          : Monthly (e.g., 2024-06)
            - 'W'          : Weekly (e.g., 2024-06-17/2024-06-23)
            - 'D'          : Daily (e.g., 2024-06-18)
            - 'H'          : Hourly (e.g., 2024-06-18 15:00)
            - 'T', 'min'   : Minutely (e.g., 2024-06-18 15:30)
            - 'S'          : Secondly (e.g., 2024-06-18 15:30:00)
        For a complete list, see:
        https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects
    case_id_col : str, default 'case:concept:name'
        The name of the column identifying each case.
    timestamp_col : str, default 'time:timestamp'
        The name of the column containing the event timestamps (must be pandas datetime dtype).

    Returns
    -------
    pd.DataFrame
        A pivot table where each row represents a case, columns represent time bins (according
        to the frequency), and values are the counts of events in each bin.

    Example
    -------
    get_pivot_timestamp_distribution(df, frequency_alias='D')
    """
    # Create column for binning
    dataframe["@@binning"] = dataframe[timestamp_col].dt.to_period(frequency_alias).astype(str)

    # Pivot table: cases as rows, yearmonths as columns, counts as values
    pivot = dataframe.pivot_table(
        index=case_id_col,
        columns="@@binning",
        values=timestamp_col,
        aggfunc="count",
        fill_value=0
    )

    pivot = pivot.reindex(sorted(pivot.columns), axis=1).reset_index()
    pivot.columns = [x if x == case_id_col else "@@evcount_"+x for x in pivot.columns]

    return pivot



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.log.obj import EventLog


def insert_event_index_as_event_attribute(
    stream, event_index_attr_name="@@eventindex"
):
    """
    Insert the current event index as event attribute

    Parameters
    -----------
    stream
        Stream
    event_index_attr_name
        Attribute name given to the event index
    """

    if not type(stream) is EventLog:
        for i in range(0, len(stream._list)):
            stream._list[i][event_index_attr_name] = i + 1

    return stream


def insert_trace_index_as_event_attribute(
    log, trace_index_attr_name="@@traceindex"
):
    """
    Inserts the current trace index as event attribute
    (overrides previous values if needed)

    Parameters
    -----------
    log
        Log
    trace_index_attr_name
        Attribute name given to the trace index
    """
    for i in range(len(log._list)):
        for j in range(len(log._list[i])):
            log._list[i][j][trace_index_attr_name] = i + 1

    return log



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util.constants import CASE_CONCEPT_NAME
from pm4py.util.constants import PARAMETER_CONSTANT_CASEID_KEY, PARAMETER_CONSTANT_ACTIVITY_KEY
from enum import Enum
from pm4py.util import exec_utils, constants, xes_constants
from typing import Optional, Dict, Any, Union, List
import polars as pl


class Parameters(Enum):
    CASE_ID_KEY = PARAMETER_CONSTANT_CASEID_KEY
    ACTIVITY_KEY = PARAMETER_CONSTANT_ACTIVITY_KEY
    POSITIVE = "positive"


def apply(
    df: pl.LazyFrame,
    prefixes: List[List[str]],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Apply a filter on prefixes of the event log

    Parameters
    -----------
    df
        LazyFrame
    prefixes
        List of prefixes to filter on
    parameters
        Parameters of the algorithm, including:
            Parameters.CASE_ID_KEY -> Column that contains the Case ID
            Parameters.ACTIVITY_KEY -> Column that contains the activity
            Parameters.POSITIVE -> Specifies if filter should be applied including (True) or excluding (False) traces

    Returns
    -----------
    df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    positive = exec_utils.get_param_value(Parameters.POSITIVE, parameters, True)

    # Create case variants
    variants_df = (
        df.sort([case_id_glue, "time:timestamp"])
        .group_by(case_id_glue, maintain_order=True)
        .agg(pl.col(activity_key).alias("activities"))
    )

    # Check if any prefix matches
    prefix_conditions = []
    for prefix in prefixes:
        prefix_list = list(prefix)
        prefix_len = len(prefix_list)
        if prefix_len == 0:
            condition = pl.lit(True)
        else:
            condition = (
                pl.col("activities").list.slice(0, prefix_len)
                == pl.lit(prefix_list)
            )
        prefix_conditions.append(condition)
    
    if prefix_conditions:
        combined_condition = prefix_conditions[0]
        for cond in prefix_conditions[1:]:
            combined_condition = combined_condition | cond
        
        matching_cases = variants_df.filter(combined_condition).select(case_id_glue)
    else:
        matching_cases = pl.DataFrame({case_id_glue: []}).lazy()

    if positive:
        ret = df.join(matching_cases, on=case_id_glue, how="inner")
    else:
        ret = df.join(matching_cases, on=case_id_glue, how="anti")

    return ret



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.ocel.util import (
    attributes_names,
    attributes_per_type,
    convergence_divergence_diagnostics,
    e2o_qualification,
    ev_att_to_obj_type,
    event_prefix_suffix_per_obj,
    events_per_object_type,
    events_per_type_per_activity,
    explode,
    extended_table,
    filtering_utils,
    flattening,
    log_ocel,
    names_stripping,
    objects_per_type_per_activity,
    ocel_consistency,
    ocel_iterator,
    ocel_to_dict_types_rel,
    ocel_type_renaming,
    parent_children_ref,
    related_events,
    related_objects,
    rename_objs_ot_tim_lex,
    sampling
)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import List, Optional, Tuple, TypeVar, Dict, Any

from pm4py.algo.discovery.inductive.cuts.abc import Cut
from pm4py.algo.discovery.inductive.cuts.concurrency import (
    ConcurrencyCutUVCL,
    ConcurrencyCutDFG,
)
from pm4py.algo.discovery.inductive.cuts.loop import LoopCutUVCL, LoopCutDFG
from pm4py.algo.discovery.inductive.cuts.sequence import (
    StrictSequenceCutUVCL,
    StrictSequenceCutDFG,
    SequenceCutUVCL,
    SequenceCutDFG,
)
from pm4py.algo.discovery.inductive.cuts.xor import (
    ExclusiveChoiceCutUVCL,
    ExclusiveChoiceCutDFG,
)
from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructure,
    IMDataStructureUVCL,
    IMDataStructureDFG,
)
from pm4py.algo.discovery.inductive.variants.instances import IMInstance
from pm4py.objects.process_tree.obj import ProcessTree
from pm4py.util import exec_utils
from enum import Enum


T = TypeVar("T", bound=IMDataStructure)
S = TypeVar("S", bound=Cut)


class Parameters(Enum):
    DISABLE_STRICT_SEQUENCE_CUT = "disable_strict_sequence_cut"


class CutFactory:

    @classmethod
    def get_cuts(
        cls,
        obj: T,
        inst: IMInstance,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[S]:
        if parameters is None:
            parameters = {}

        disable_strict_sequence_cut = exec_utils.get_param_value(
            Parameters.DISABLE_STRICT_SEQUENCE_CUT, parameters, False
        )

        if inst is IMInstance.IM or inst is IMInstance.IMf:
            if type(obj) is IMDataStructureUVCL:
                sequence_cut = StrictSequenceCutUVCL
                if disable_strict_sequence_cut:
                    sequence_cut = SequenceCutUVCL
                return [
                    ExclusiveChoiceCutUVCL,
                    sequence_cut,
                    ConcurrencyCutUVCL,
                    LoopCutUVCL,
                ]
        if inst is IMInstance.IMd:
            if type(obj) is IMDataStructureDFG:
                sequence_cut = StrictSequenceCutDFG
                if disable_strict_sequence_cut:
                    sequence_cut = SequenceCutDFG
                return [
                    ExclusiveChoiceCutDFG,
                    sequence_cut,
                    ConcurrencyCutDFG,
                    LoopCutDFG,
                ]
        return list()

    @classmethod
    def find_cut(
        cls,
        obj: IMDataStructure,
        inst: IMInstance,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Optional[Tuple[ProcessTree, List[T]]]:
        for c in CutFactory.get_cuts(obj, inst, parameters):
            r = c.apply(obj, parameters)
            if r is not None:
                return r
        return None



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any

from pm4py.streaming.algo.conformance.temporal.variants import classic
from pm4py.util import exec_utils
from pm4py.util import typing


class Variants(Enum):
    CLASSIC = classic


def apply(
    temporal_profile: typing.TemporalProfile,
    variant=Variants.CLASSIC,
    parameters: Optional[Dict[Any, Any]] = None,
):
    """
    Initialize the streaming conformance checking

    Parameters
    ---------------
    temporal_profile
        Temporal profile
    variant
        Variant of the algorithm, possible values:
        - Variants.CLASSIC
    parameters
        Parameters of the algorithm, including:
         - Parameters.ACTIVITY_KEY => the attribute to use as activity
         - Parameters.START_TIMESTAMP_KEY => the attribute to use as start timestamp
         - Parameters.TIMESTAMP_KEY => the attribute to use as timestamp
         - Parameters.ZETA => multiplier for the standard deviation
         - Parameters.CASE_ID_KEY => column to use as case identifier
         - Parameters.DICT_VARIANT => the variant of dictionary to use
         - Parameters.CASE_DICT_ID => the identifier of the case dictionary
         - Parameters.DEV_DICT_ID => the identifier of the deviations dictionary
    """
    return exec_utils.get_variant(variant).apply(
        temporal_profile, parameters=parameters
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.log.util import sorting
from pm4py.objects.log.util import basic_filter
from pm4py.util import points_subset
from pm4py.util import xes_constants as xes
from pm4py.util import exec_utils
from enum import Enum
from pm4py.util import constants
from typing import Optional, Dict, Any, Union, List
from pm4py.objects.log.obj import EventLog


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    ATTRIBUTE_KEY = constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY
    PARAMETER_SAMPLE_SIZE = "sample_size"
    SORT_LOG_REQUIRED = "sort_log_required"


def apply(
    log: EventLog,
    list_activities: List[str],
    sample_size: int,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Dict[str, Any]:
    """
    Finds the performance spectrum provided a log
    and a list of activities

    Parameters
    -------------
    log
        Log
    list_activities
        List of activities interesting for the performance spectrum (at least two)
    sample_size
        Size of the sample
    parameters
        Parameters of the algorithm,  including:
            - Parameters.ACTIVITY_KEY
            - Parameters.TIMESTAMP_KEY

    Returns
    -------------
    points
        Points of the performance spectrum
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes.DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, xes.DEFAULT_TIMESTAMP_KEY
    )
    sort_log_required = exec_utils.get_param_value(
        Parameters.SORT_LOG_REQUIRED, parameters, True
    )

    parameters[Parameters.ATTRIBUTE_KEY] = activity_key
    log = basic_filter.filter_log_events_attr(
        log, list_activities, parameters=parameters
    )
    if sort_log_required:
        log = sorting.sort_timestamp_log(log, timestamp_key=timestamp_key)

    points = []

    for trace in log:
        for i in range(len(trace) - len(list_activities) + 1):
            acti_comb = [
                event[activity_key]
                for event in trace[i: i + len(list_activities)]
            ]

            if acti_comb == list_activities:
                timest_comb = [
                    event[timestamp_key].timestamp()
                    for event in trace[i: i + len(list_activities)]
                ]

                points.append(timest_comb)

    points = sorted(points, key=lambda x: x[0])

    if len(points) > sample_size:
        points = points_subset.pick_chosen_points_list(sample_size, points)

    return points



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from graphviz import Digraph
from enum import Enum

import pm4py
from pm4py.util import exec_utils, constants, xes_constants
from typing import Optional, Dict, Any
import pandas as pd
from uuid import uuid4
from pm4py.util import vis_utils
import tempfile
from pm4py.algo.filtering.dfg import dfg_filtering


def __get_freq_perf_df(
    dataframe: pd.DataFrame,
    activity_key: str,
    aggregation_measure: str,
    activity_percentage: float,
    paths_percentage: float,
    dependency_threshold: float,
):
    """
    Gets the frequency and performance DFG abstractions from the provided dataframe
    (internal usage)
    """
    freq_dfg, sa, ea = pm4py.discover_dfg(dataframe)
    perf_dfg, sa, ea = pm4py.discover_performance_dfg(dataframe)
    act_count = pm4py.get_event_attribute_values(dataframe, activity_key)

    freq_dfg, sa, ea, act_count = (
        dfg_filtering.filter_dfg_on_activities_percentage(
            freq_dfg, sa, ea, act_count, activity_percentage
        )
    )
    freq_dfg, sa, ea, act_count = dfg_filtering.filter_dfg_on_paths_percentage(
        freq_dfg, sa, ea, act_count, paths_percentage
    )
    freq_dfg, sa, ea, act_count = dfg_filtering.filter_dfg_keep_connected(
        freq_dfg, sa, ea, act_count, dependency_threshold
    )

    perf_dfg = {
        x: y[aggregation_measure] for x, y in perf_dfg.items() if x in freq_dfg
    }

    return freq_dfg, perf_dfg, sa, ea, act_count


class Parameters(Enum):
    FORMAT = "format"
    BGCOLOR = "bgcolor"
    RANKDIR = "rankdir"
    ANNOTATION = "annotation"
    AGGREGATION_MEASURE = "aggregation_measure"
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    ACTIVITY_PERCENTAGE = "activity_percentage"
    PATHS_PERCENTAGE = "paths_percentage"
    DEPENDENCY_THRESHOLD = "dependency_threshold"
    MIN_FACT_EDGES_INTERLEAVINGS = "min_fact_edges_interleavings"
    ENABLE_GRAPH_TITLE = "enable_graph_title"
    GRAPH_TITLE = "graph_title"


def apply(
    dataframe1: pd.DataFrame,
    dataframe2: pd.DataFrame,
    interleavings: pd.DataFrame,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Digraph:
    """
    Visualizes the interleavings discovered between two different processes.
    We suppose to provide both event logs, and the discovered interleavings.
    The visualization includes the DFG of both processes, along with the arcs discovered between them.
    Both frequency and performance visualization are available.

    Parameters
    --------------------
    dataframe1
        Dataframe of the first process
    dataframe2
        Dataframe of the second process
    interleavings
        Interleavings between the two considered processes
    parameters
        Parameters of the algorithm, including:
        - Parameters.FORMAT => the format of the visualization
        - Parameters.BGCOLOR => the background color
        - Parameters.RANKDIR => the rank direction (LR or TB; default: TB)
        - Parameters.ANNOTATION => the annotation to represent (possible values: frequency or performance)
        - Parameters.AGGREGATION_MEASURE => which aggregation should be used when considering performance
        - Parameters.ACTIVITY_KEY => the activity key
        - Parameters.ACTIVITY_PERCENTAGE => the percentage of activities to include for the DFG of the single processes
        - Parameters.PATHS_PERCENTAGE => the percentage of paths to include for the DFG of the single processes
        - Parameters.DEPENDENCY_THRESHOLD => the dependency threshold to consider for the DFG of the single processes
        - Parameters.MIN_FACT_EDGES_INTERLEAVINGS => factor that is multiplied to the minimum number of occurrences of
                                        edges in the single processes, to decide if the interleavings edge should
                                        be included. E.g., if 0.3 is provided, only interleavings edges having a frequency
                                        of at least 0.3 * MIN_EDGE_COUNT_IN_PROCESSES are included.
    Returns
    ----------------
    digraph
        Graphviz Digraph
    """
    if parameters is None:
        parameters = {}

    image_format = exec_utils.get_param_value(
        Parameters.FORMAT, parameters, "png"
    )
    bgcolor = exec_utils.get_param_value(
        Parameters.BGCOLOR, parameters, constants.DEFAULT_BGCOLOR
    )
    rankdir = exec_utils.get_param_value(
        Parameters.RANKDIR, parameters, constants.DEFAULT_RANKDIR_GVIZ
    )
    annotation = exec_utils.get_param_value(
        Parameters.ANNOTATION, parameters, "frequency"
    )
    aggregation_measure = exec_utils.get_param_value(
        Parameters.AGGREGATION_MEASURE, parameters, "mean"
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    activity_percentage = exec_utils.get_param_value(
        Parameters.ACTIVITY_PERCENTAGE, parameters, 0.3
    )
    paths_percentage = exec_utils.get_param_value(
        Parameters.PATHS_PERCENTAGE, parameters, 0.3
    )
    dependency_threshold = exec_utils.get_param_value(
        Parameters.DEPENDENCY_THRESHOLD, parameters, 0.3
    )
    min_fact_edges_interleavings = exec_utils.get_param_value(
        Parameters.MIN_FACT_EDGES_INTERLEAVINGS, parameters, 0.3
    )

    enable_graph_title = exec_utils.get_param_value(
        Parameters.ENABLE_GRAPH_TITLE,
        parameters,
        constants.DEFAULT_ENABLE_GRAPH_TITLES,
    )
    graph_title = exec_utils.get_param_value(
        Parameters.GRAPH_TITLE, parameters, "Interleavings"
    )

    filename = tempfile.NamedTemporaryFile(suffix=".gv")
    filename.close()

    viz = Digraph(
        "interleavings",
        filename=filename.name,
        engine="dot",
        graph_attr={"bgcolor": bgcolor},
    )

    if enable_graph_title:
        viz.attr(
            label='<<FONT POINT-SIZE="20">' + graph_title + "</FONT>>",
            labelloc="top",
        )

    viz.attr("node", shape="ellipse", fixedsize="false")
    viz.attr(rankdir=rankdir)
    viz.format = image_format.replace("html", "plain-ext")

    freq_dfg1, perf_dfg1, sa1, ea1, act_count1 = __get_freq_perf_df(
        dataframe1,
        activity_key,
        aggregation_measure,
        activity_percentage,
        paths_percentage,
        dependency_threshold,
    )
    freq_dfg2, perf_dfg2, sa2, ea2, act_count2 = __get_freq_perf_df(
        dataframe2,
        activity_key,
        aggregation_measure,
        activity_percentage,
        paths_percentage,
        dependency_threshold,
    )

    min_act_count = min(min(act_count1.values()), min(act_count2.values()))
    max_act_count = max(max(act_count1.values()), max(act_count2.values()))

    min_edge_count = min([min(freq_dfg1.values()), min(freq_dfg2.values())])

    interleavings_lr_frequency = (
        interleavings[interleavings["@@direction"] == "LR"][
            ["@@source_activity", "@@target_activity"]
        ]
        .value_counts()
        .to_dict()
    )
    interleavings_lr_performance = (
        interleavings[interleavings["@@direction"] == "LR"]
        .groupby(["@@source_activity", "@@target_activity"])[
            "@@timestamp_diff"
        ]
        .agg(aggregation_measure)
        .to_dict()
    )
    interleavings_rl_frequency = (
        interleavings[interleavings["@@direction"] == "RL"][
            ["@@source_activity", "@@target_activity"]
        ]
        .value_counts()
        .to_dict()
    )
    interleavings_rl_performance = (
        interleavings[interleavings["@@direction"] == "RL"]
        .groupby(["@@source_activity", "@@target_activity"])[
            "@@timestamp_diff"
        ]
        .agg(aggregation_measure)
        .to_dict()
    )

    interleavings_lr_frequency = {
        x: y
        for x, y in interleavings_lr_frequency.items()
        if x[0] in act_count1
        and x[1] in act_count2
        and y >= min_edge_count * min_fact_edges_interleavings
    }
    interleavings_rl_frequency = {
        x: y
        for x, y in interleavings_rl_frequency.items()
        if x[0] in act_count2
        and x[1] in act_count1
        and y >= min_edge_count * min_fact_edges_interleavings
    }
    interleavings_lr_performance = {
        x: y
        for x, y in interleavings_lr_performance.items()
        if x[0] in act_count1
        and x[1] in act_count2
        and x in interleavings_lr_frequency
    }
    interleavings_rl_performance = {
        x: y
        for x, y in interleavings_rl_performance.items()
        if x[0] in act_count2
        and x[1] in act_count1
        and x in interleavings_rl_frequency
    }

    min_edge_count = min(
        [
            min(freq_dfg1.values()),
            min(freq_dfg2.values()),
            min(interleavings_lr_frequency.values()),
            min(interleavings_rl_frequency.values()),
            min(sa1.values()),
            min(sa2.values()),
            min(ea1.values()),
            min(ea2.values()),
        ]
    )
    max_edge_count = max(
        [
            max(freq_dfg1.values()),
            max(freq_dfg2.values()),
            max(interleavings_lr_frequency.values()),
            max(interleavings_rl_frequency.values()),
            max(sa1.values()),
            max(sa2.values()),
            max(ea1.values()),
            max(ea2.values()),
        ]
    )

    min_edge_perf = min(
        [
            min(perf_dfg1.values()),
            min(perf_dfg2.values()),
            min(interleavings_lr_performance.values()),
            min(interleavings_rl_performance.values()),
        ]
    )
    max_edge_perf = max(
        [
            max(perf_dfg1.values()),
            max(perf_dfg2.values()),
            max(interleavings_lr_performance.values()),
            max(interleavings_rl_performance.values()),
        ]
    )

    nodes1 = {}
    nodes2 = {}

    with viz.subgraph(name="First Model") as c1:
        c1.attr(style="filled")
        c1.attr(color="lightgray")
        c1.attr(label="First Model")

        for act in act_count1:
            act_uuid = str(uuid4())
            nodes1[act] = act_uuid
            color = vis_utils.get_trans_freq_color(
                act_count1[act], min_act_count, max_act_count
            )
            c1.node(
                act_uuid,
                label=act + "\n" + str(act_count1[act]),
                shape="box",
                style="filled",
                fillcolor=color,
            )

        for edge in freq_dfg1:
            if annotation == "frequency":
                count = freq_dfg1[edge]
                label = str(count)
                penwidth = str(
                    vis_utils.get_arc_penwidth(
                        count, min_edge_count, max_edge_count
                    )
                )
            elif annotation == "performance":
                perf = perf_dfg1[edge]
                label = vis_utils.human_readable_stat(perf)
                penwidth = str(
                    vis_utils.get_arc_penwidth(
                        perf, min_edge_perf, max_edge_perf
                    )
                )
            viz.edge(
                nodes1[edge[0]],
                nodes1[edge[1]],
                label=label,
                penwidth=penwidth,
            )

        c1.node(
            "@@startnode1",
            constants.DEFAULT_START_SYMBOL_GRAPHS,
            shape="circle",
            fontsize="34",
            color="black",
            fontcolor="black",
        )
        c1.node(
            "@@endnode1",
            constants.DEFAULT_END_SYMBOL_GRAPHS,
            shape="doublecircle",
            fontsize="32",
            color="black",
            fontcolor="black",
        )

        for sa in sa1:
            penwidth = str(
                vis_utils.get_arc_penwidth(
                    sa1[sa], min_edge_count, max_edge_count
                )
            )
            label = str(sa1[sa]) if annotation == "frequency" else " "
            viz.edge(
                "@@startnode1",
                nodes1[sa],
                color="black",
                label=label,
                penwidth=penwidth,
            )

        for ea in ea1:
            penwidth = str(
                vis_utils.get_arc_penwidth(
                    ea1[ea], min_edge_count, max_edge_count
                )
            )
            label = str(ea1[ea]) if annotation == "frequency" else " "
            viz.edge(
                nodes1[ea],
                "@@endnode1",
                color="black",
                label=label,
                penwidth=penwidth,
            )

    with viz.subgraph(name="Second Model") as c2:
        c2.attr(style="filled")
        c2.attr(color="lightgray")
        c2.attr(label="Second Model")

        for act in act_count2:
            act_uuid = str(uuid4())
            nodes2[act] = act_uuid
            color = vis_utils.get_trans_freq_color(
                act_count2[act], min_act_count, max_act_count
            )
            c2.node(
                act_uuid,
                label=act + "\n" + str(act_count2[act]),
                shape="box",
                style="filled",
                fillcolor=color,
                color="gray",
                fontcolor="gray",
            )

        for edge in freq_dfg2:
            if annotation == "frequency":
                count = freq_dfg2[edge]
                label = str(count)
                penwidth = str(
                    vis_utils.get_arc_penwidth(
                        count, min_edge_count, max_edge_count
                    )
                )
            elif annotation == "performance":
                perf = perf_dfg2[edge]
                label = vis_utils.human_readable_stat(perf)
                penwidth = str(
                    vis_utils.get_arc_penwidth(
                        perf, min_edge_perf, max_edge_perf
                    )
                )
            viz.edge(
                nodes2[edge[0]],
                nodes2[edge[1]],
                label=label,
                penwidth=penwidth,
                color="gray",
                fontcolor="gray",
            )

        c2.node(
            "@@startnode2",
            constants.DEFAULT_START_SYMBOL_GRAPHS,
            shape="circle",
            fontsize="34",
            color="gray",
            fontcolor="gray",
        )
        c2.node(
            "@@endnode2",
            constants.DEFAULT_END_SYMBOL_GRAPHS,
            shape="doublecircle",
            fontsize="32",
            color="gray",
            fontcolor="gray",
        )

        for sa in sa2:
            penwidth = str(
                vis_utils.get_arc_penwidth(
                    sa2[sa], min_edge_count, max_edge_count
                )
            )
            label = str(sa2[sa]) if annotation == "frequency" else " "
            viz.edge(
                "@@startnode2",
                nodes2[sa],
                color="gray",
                label=label,
                penwidth=penwidth,
            )

        for ea in ea2:
            penwidth = str(
                vis_utils.get_arc_penwidth(
                    ea2[ea], min_edge_count, max_edge_count
                )
            )
            label = str(ea2[ea]) if annotation == "frequency" else " "
            viz.edge(
                nodes2[ea],
                "@@endnode2",
                color="gray",
                label=label,
                penwidth=penwidth,
            )

    for edge in interleavings_lr_frequency:
        if annotation == "frequency":
            count = interleavings_lr_frequency[edge]
            label = str(count)
            penwidth = str(
                vis_utils.get_arc_penwidth(
                    count, min_edge_count, max_edge_count
                )
            )
        elif annotation == "performance":
            perf = interleavings_lr_performance[edge]
            label = vis_utils.human_readable_stat(perf)
            penwidth = str(
                vis_utils.get_arc_penwidth(perf, min_edge_perf, max_edge_perf)
            )
        viz.edge(
            nodes1[edge[0]],
            nodes2[edge[1]],
            label=label,
            penwidth=penwidth,
            color="violet",
            fontcolor="violet",
            style="dashed",
        )

    for edge in interleavings_rl_frequency:
        if annotation == "frequency":
            count = interleavings_rl_frequency[edge]
            label = str(count)
            penwidth = str(
                vis_utils.get_arc_penwidth(
                    count, min_edge_count, max_edge_count
                )
            )
        elif annotation == "performance":
            perf = interleavings_rl_frequency[edge]
            label = vis_utils.human_readable_stat(perf)
            penwidth = str(
                vis_utils.get_arc_penwidth(perf, min_edge_perf, max_edge_perf)
            )
        viz.edge(
            nodes2[edge[0]],
            nodes1[edge[1]],
            label=label,
            penwidth=penwidth,
            color="violet",
            fontcolor="violet",
            style="dashed",
        )

    return viz



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any

from copy import deepcopy
from pm4py.objects.ocel.obj import OCEL
from pm4py.util import exec_utils


class Parameters(Enum):
    EXCLUDE_OBJECT_TYPES = "exclude_object_types"


def apply(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None) -> OCEL:
    """
    Rename objects given their object type, lifecycle start/end timestamps, and lexicographic order,

    Parameters
    -----------------
    ocel
        Object-centric event log
    parameters
        Parameters of the algorithm, including:
        - Parameters.EXCLUDE_OBJECT_TYPES => set/list of object types to exclude from renaming

    Returns
    ----------------
    renamed_ocel
        Object-centric event log with renaming
    """
    if parameters is None:
        parameters = {}

    exclude_object_types = exec_utils.get_param_value(
        Parameters.EXCLUDE_OBJECT_TYPES, parameters, set()
    )
    # normalize to a Python set for fast membership checks
    try:
        exclude_object_types = set(exclude_object_types)
    except Exception:
        exclude_object_types = set()

    objects_start = (
        ocel.relations.groupby(ocel.object_id_column)[ocel.event_timestamp]
        .first()
        .to_dict()
    )
    objects_end = (
        ocel.relations.groupby(ocel.object_id_column)[ocel.event_timestamp]
        .last()
        .to_dict()
    )
    objects_ot0 = ocel.objects[
        [ocel.object_id_column, ocel.object_type_column]
    ].to_dict("records")
    objects_ot0 = [
        (x[ocel.object_id_column], x[ocel.object_type_column])
        for x in objects_ot0
    ]
    objects_ot1 = {}

    for el in objects_ot0:
        if not el[1] in objects_ot1:
            objects_ot1[el[1]] = []
        objects_ot1[el[1]].append(el[0])

    overall_objects = {}
    keys = sorted(list(objects_ot1))
    for ot in keys:
        objects = objects_ot1[ot]
        objects.sort(key=lambda x: (objects_start[x], objects_end[x], x))
        if ot in exclude_object_types:
            # skip renaming for this object type
            continue
        mapped = {objects[i]: ot + "_" + str(i + 1) for i in range(len(objects))}
        overall_objects.update(mapped)

    ocel = deepcopy(ocel)
    # map and keep original IDs for excluded object types
    oid_col = ocel.object_id_column
    ocel.objects[oid_col] = ocel.objects[oid_col].map(overall_objects).fillna(
        ocel.objects[oid_col]
    )
    ocel.relations[oid_col] = ocel.relations[oid_col].map(overall_objects).fillna(
        ocel.relations[oid_col]
    )
    if ocel.o2o is not None:
        ocel.o2o[oid_col] = ocel.o2o[oid_col].map(overall_objects).fillna(
            ocel.o2o[oid_col]
        )
        ocel.o2o[oid_col + "_2"] = ocel.o2o[oid_col + "_2"].map(overall_objects).fillna(
            ocel.o2o[oid_col + "_2"]
        )
    if ocel.object_changes is not None:
        ocel.object_changes[oid_col] = ocel.object_changes[oid_col].map(
            overall_objects
        ).fillna(ocel.object_changes[oid_col])

    return ocel
